__author__ = "aiXplain"

"""
Copyright 2023 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: aiXplain team
Date: March 20th 2023
Description:
    Function Enum
"""

import logging

from aixplain.utils import config
from aixplain.utils.request_utils import _request_with_retry
from enum import Enum
from urllib.parse import urljoin


def load_functions():
    api_key = config.TEAM_API_KEY
    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    url = urljoin(backend_url, "sdk/functions")
    if aixplain_key != "":
        api_key = aixplain_key
        headers = {"x-aixplain-key": aixplain_key, "Content-Type": "application/json"}
    else:
        headers = {"x-api-key": api_key, "Content-Type": "application/json"}
    r = _request_with_retry("get", url, headers=headers)
    if not 200 <= r.status_code < 300:
        raise Exception(
            f'Functions could not be loaded, probably due to the set API key (e.g. "{api_key}") is not valid. For help, please refer to the documentation (https://github.com/aixplain/aixplain#api-key-setup)'
        )
    resp = r.json()
    functions = Enum("Function", {w["id"].upper().replace("-", "_"): w["id"] for w in resp["items"]}, type=str)
    functions_input_output = {
        function["id"]: {
            "input": {
                input_data_object["dataType"]
                for input_data_object in function["params"]
                if input_data_object["required"] is True
            },
            "output": {output_data_object["dataType"] for output_data_object in function["output"]},
            "spec": function
        }
        for function in resp["items"]
    }
    return functions, functions_input_output


Function, FunctionInputOutput = load_functions()



__author__ = "shreyassharma"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: March 27th 2023
Description:
    Corpus Factory Class
"""

import aixplain.utils.config as config
import aixplain.processes.data_onboarding.onboard_functions as onboard_functions
import json
import logging
import shutil

from aixplain.factories.asset_factory import AssetFactory
from aixplain.factories.data_factory import DataFactory
from aixplain.modules.corpus import Corpus
from aixplain.modules.data import Data
from aixplain.modules.metadata import MetaData
from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.error_handler import ErrorHandler
from aixplain.enums.function import Function
from aixplain.enums.language import Language
from aixplain.enums.license import License
from aixplain.enums.privacy import Privacy
from aixplain.utils.file_utils import _request_with_retry
from aixplain.utils import config
from pathlib import Path
from tqdm import tqdm
from typing import Any, Dict, List, Optional, Text, Union
from urllib.parse import urljoin
from warnings import warn


class CorpusFactory(AssetFactory):
    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def __from_response(cls, response: Dict) -> Corpus:
        """Converts Json response to 'Corpus' object

        Args:
            response (dict): Json from API

        Returns:
            Dataset: Coverted 'Dataset' object
        """
        data = []
        for d in response["data"]:
            languages = []
            if "languages" in d["metadata"]:
                languages = []
                for lng in d["metadata"]["languages"]:
                    if "dialect" not in lng:
                        lng["dialect"] = ""
                    languages.append(Language(lng))
            data.append(
                Data(
                    id=d["id"],
                    name=d["name"],
                    dtype=DataType(d["dataType"]),
                    dsubtype=DataSubtype(d["dataSubtype"]),
                    privacy=Privacy.PRIVATE,
                    languages=languages,
                    onboard_status=d["status"],
                )
            )
        functions = [Function(f) for f in response["suggestedFunction"]]

        try:
            license = License(response["license"]["typeId"])
        except:
            license = None

        try:
            length = int(response["segmentsCount"])
        except:
            length = None

        corpus = Corpus(
            id=response["id"],
            name=response["name"],
            description=response["description"],
            functions=functions,
            license=license,
            data=data,
            onboard_status=response["status"],
            length=length,
        )
        return corpus

    @classmethod
    def get(cls, corpus_id: Text) -> Corpus:
        """Create a 'Corpus' object from corpus id

        Args:
            corpus_id (Text): Corpus ID of required corpus.

        Returns:
            Corpus: Created 'Corpus' object
        """
        url = urljoin(cls.backend_url, f"sdk/corpora/{corpus_id}/overview")
        if cls.aixplain_key != "":
            headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        logging.info(f"Start service for GET Corpus  - {url} - {headers}")
        r = _request_with_retry("get", url, headers=headers)
        resp = r.json()
        if "statusCode" in resp and resp["statusCode"] == 404:
            raise Exception(f"Corpus GET Error: Dataset {corpus_id} not found.")
        return cls.__from_response(resp)

    @classmethod
    def create_asset_from_id(cls, corpus_id: Text) -> Corpus:
        warn(
            'This method will be deprecated in the next versions of the SDK. Use "get" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.get(corpus_id)

    @classmethod
    def list(
        cls,
        query: Optional[Text] = None,
        function: Optional[Function] = None,
        language: Optional[Union[Language, List[Language]]] = None,
        data_type: Optional[DataType] = None,
        license: Optional[License] = None,
        page_number: int = 0,
        page_size: int = 20,
    ) -> Dict:
        """Corpus Listing

        Args:
            query (Optional[Text], optional): search query. Defaults to None.
            function (Optional[Function], optional): function filter. Defaults to None.
            language (Optional[Union[Language, List[Language]]], optional): language filter. Defaults to None.
            data_type (Optional[DataType], optional): data type filter. Defaults to None.
            license (Optional[License], optional): license filter. Defaults to None.
            page_number (int, optional): page number. Defaults to 0.
            page_size (int, optional): page size. Defaults to 20.

        Returns:
            Dict: list of corpora in agreement with the filters, page number, page total and total elements
        """
        url = urljoin(cls.backend_url, "sdk/corpora/paginate")
        if cls.aixplain_key != "":
            headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}

        assert 0 < page_size <= 100, f"Corpus List Error: Page size must be greater than 0 and not exceed 100."
        payload = {"pageSize": page_size, "pageNumber": page_number, "sort": [{"field": "createdAt", "dir": -1}]}

        if query is not None:
            payload["q"] = str(query)

        if function is not None:
            payload["function"] = function.value

        if license is not None:
            payload["license"] = license.value

        if data_type is not None:
            payload["dataType"] = data_type.value

        if language is not None:
            if isinstance(language, Language):
                language = [language]
            payload["language"] = [lng.value["language"] for lng in language]

        logging.info(f"Start service for POST List Corpus - {url} - {headers} - {json.dumps(payload)}")
        r = _request_with_retry("post", url, headers=headers, json=payload)
        resp = r.json()
        corpora, page_total, total = [], 0, 0
        if "results" in resp:
            results = resp["results"]
            page_total = resp["pageTotal"]
            total = resp["total"]
            logging.info(f"Response for POST List Corpus - Page Total: {page_total} / Total: {total}")
            for corpus in results:
                corpus_ = cls.__from_response(corpus)
                # add languages
                languages = []
                for lng in corpus["languages"]:
                    if "dialect" not in lng:
                        lng["dialect"] = ""
                    languages.append(Language(lng))
                corpus_.kwargs["languages"] = languages
                corpora.append(corpus_)
        return {"results": corpora, "page_total": page_total, "page_number": page_number, "total": total}

    @classmethod
    def get_assets_from_page(
        cls, page_number: int = 1, task: Optional[Function] = None, language: Optional[Text] = None
    ) -> List[Corpus]:
        """Get the list of corpora from a given page. Additional task and language filters can be also be provided

        Args:
            page_number (int, optional): Page from which corpora are to be listed. Defaults to 1.
            task (Function, optional): Task of listed corpora. Defaults to None.
            language (Text, optional): language of listed corpora. Defaults to None.

        Returns:
            List[Corpus]: List of corpora based on given filters
        """
        warn(
            'This method will be deprecated in the next versions of the SDK. Use "list" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        if language is not None:
            language = [language]
        return cls.list(function=task, page_number=page_number, languages=language)

    @classmethod
    def create(
        cls,
        name: Text,
        description: Text,
        license: License,
        content_path: Union[Union[Text, Path], List[Union[Text, Path]]],
        schema: List[Union[Dict, MetaData]],
        ref_data: List[Any] = [],
        tags: List[Text] = [],
        functions: List[Function] = [],
        privacy: Privacy = Privacy.PRIVATE,
        error_handler: ErrorHandler = ErrorHandler.SKIP,
        api_key: Optional[Text] = None
    ) -> Dict:
        """Asynchronous call to Upload a corpus to the user's dashboard.

        Args:
            name (Text): corpus name
            description (Text): corpus description
            license (License): corpus license
            content_path (Union[Union[Text, Path], List[Union[Text, Path]]]): path to .csv files containing the data
            schema (List[Union[Dict, MetaData]]): meta data
            ref_data (Optional[List[Union[Text, Data]]], optional): referencing data which already exists and should be part of the corpus. Defaults to [].
            tags (Optional[List[Text]], optional): tags that explain the corpus. Defaults to [].
            functions (Optional[List[Function]], optional): AI functions for which the corpus may be used. Defaults to [].
            privacy (Optional[Privacy], optional): visibility of the corpus. Defaults to Privacy.PRIVATE.
            error_handler (ErrorHandler, optional): how to handle failed rows in the data asset. Defaults to ErrorHandler.SKIP.
            api_key (Optional[Text]): team api key. Defaults to None.

        Returns:
            Dict: response dict
        """
        folder, return_dict = None, {}
        # check team key
        try:
            assert (
                len(schema) > 0 or len(ref_data) > 0
            ), "Data Asset Onboarding Error: You must specify a data to onboard a corpus."

            content_paths = content_path
            if isinstance(content_path, list) is False:
                content_paths = [content_path]

            for i, metadata in enumerate(schema):
                if isinstance(metadata, dict):
                    schema[i] = MetaData(**metadata)

            if len(ref_data) > 0:
                if isinstance(ref_data[0], Data):
                    ref_data = [w.id for w in ref_data]
                # check whether the referred data exist. Otherwise, raise an exception
                for i, data_id in enumerate(ref_data):
                    if onboard_functions.is_data(data_id) is False:
                        message = f"Data Asset Onboarding Error: Referenced Data {data_id} does not exist."
                        logging.exception(message)
                        raise Exception(message)
                    ref_data[i] = DataFactory.get(data_id=data_id)

            # check whether reserved names are used as data/column names
            for metadata in schema:
                for forbidden_name in onboard_functions.FORBIDDEN_COLUMN_NAMES:
                    if forbidden_name in [metadata.name, metadata.data_column]:
                        message = f"Data Asset Onboarding Error: {forbidden_name} is reserved name and must not be used as the name of a data or a column."
                        logging.error(message)
                        raise Exception(message)

            # get file extension paths to process
            paths = onboard_functions.get_paths(content_paths)

            # process data and create files
            folder = Path(name)
            folder.mkdir(exist_ok=True)

            dataset = []
            for i in tqdm(range(len(schema)), desc=" Corpus onboarding progress", position=0):
                metadata = schema[i]
                if metadata.privacy is None:
                    metadata.privacy = privacy

                files, data_column_idx, start_column_idx, end_column_idx, nrows = onboard_functions.process_data_files(
                    data_asset_name=name, metadata=metadata, paths=paths, folder=name
                )

                dataset.append(
                    Data(
                        id="",
                        name=metadata.name,
                        dtype=metadata.dtype,
                        dsubtype=metadata.dsubtype,
                        privacy=metadata.privacy,
                        onboard_status="onboarding",
                        data_column=data_column_idx,
                        start_column=start_column_idx,
                        end_column=end_column_idx,
                        files=files,
                        languages=metadata.languages,
                        length=nrows,
                    )
                )

            # check alignment
            sizes = [d.length for d in dataset] + [d.length for d in ref_data]
            assert (
                len(set(sizes)) == 1
            ), f"Data Asset Onboarding Error: All data must have the same number of rows. Lengths: {str(set(sizes))}"

            corpus = Corpus(
                id="",
                name=name,
                description=description,
                data=dataset,
                functions=functions,
                tags=tags,
                license=license,
                privacy=privacy,
                onboard_status="onboarding",
            )

            corpus_payload = onboard_functions.build_payload_corpus(corpus, [ref.id for ref in ref_data], error_handler)

            response = onboard_functions.create_data_asset(payload=corpus_payload, api_key=api_key)
            if response["success"] is True:
                return_dict = {"status": response["status"], "asset_id": response["asset_id"]}
            else:
                raise Exception(response["error"])
            shutil.rmtree(folder)
        except Exception as e:
            if folder is not None:
                shutil.rmtree(folder)
            raise Exception(e)
        return return_dict



__author__ = "aiXplain"

"""
Copyright 2024 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Lucas Pavanelli and Thiago Castro Ferreira
Date: May 16th 2024
Description:
    Agentification Class
"""
from typing import Text, Union

from aixplain.modules.agent.tool import Tool
from aixplain.modules.pipeline import Pipeline


class PipelineTool(Tool):
    """Specialized software or resource designed to assist the AI in executing specific tasks or functions based on user commands.

    Attributes:
        description (Text): descriptiion of the tool
        pipeline (Union[Text, Pipeline]): pipeline
    """

    def __init__(
        self,
        description: Text,
        pipeline: Union[Text, Pipeline],
        **additional_info,
    ) -> None:
        """Specialized software or resource designed to assist the AI in executing specific tasks or functions based on user commands.

        Args:
            description (Text): description of the tool
            pipeline (Union[Text, Pipeline]): pipeline
        """
        super().__init__("", description, **additional_info)
        if isinstance(pipeline, Pipeline):
            pipeline = pipeline.id
        self.pipeline = pipeline

    def validate(self):
        from aixplain.factories.pipeline_factory import PipelineFactory

        try:
            PipelineFactory.get(self.pipeline)
        except Exception:
            raise Exception(f"Pipeline Tool Unavailable. Make sure Pipeline '{self.pipeline}' exists or you have access to it.")



__author__ = "michaellam"

import pytest
import time

from aixplain.factories.model_factory import ModelFactory
from tests.test_utils import delete_asset
from aixplain.utils import config


@pytest.mark.skip(reason="Model Deployment is deactivated for improvements.")
def test_deploy_model():
    # Start the deployment
    model_name = "Test Model"
    repo_id = "tiiuae/falcon-7b"
    response = ModelFactory.deploy_huggingface_model(model_name, repo_id, hf_token=config.HF_TOKEN)
    assert "id" in response.keys()

    # Check for status
    model_id = response["id"]
    num_retries = 120
    counter = 0
    while ModelFactory.get_huggingface_model_status(model_id)["status"].lower() != "onboarded":
        time.sleep(10)
        counter += 1
        if counter == num_retries:
            assert ModelFactory.get_huggingface_model_status(model_id)["status"].lower() == "onboarded"

    # Clean up
    delete_asset(model_id, config.TEAM_API_KEY)


# @pytest.mark.skip(reason="Model Deployment is deactivated for improvements.")
def test_nonexistent_model():
    # Start the deployment
    model_name = "Test Model"
    repo_id = "nonexistent-supplier/nonexistent-model"
    response = ModelFactory.deploy_huggingface_model(model_name, repo_id, hf_token=config.HF_TOKEN)
    assert response["statusCode"] == 400
    assert response["message"] == "err.unable_to_onboard_model"


# @pytest.mark.skip(reason="Model Deployment is deactivated for improvements.")
def test_size_limit():
    # Start the deployment
    model_name = "Test Model"
    repo_id = "tiiuae/falcon-40b"
    response = ModelFactory.deploy_huggingface_model(model_name, repo_id, hf_token=config.HF_TOKEN)
    assert response["statusCode"] == 400
    assert response["message"] == "err.unable_to_onboard_model"


# @pytest.mark.skip(reason="Model Deployment is deactivated for improvements.")
def test_gated_model():
    # Start the deployment
    model_name = "Test Model"
    repo_id = "meta-llama/Llama-2-7b-hf"
    response = ModelFactory.deploy_huggingface_model(model_name, repo_id, hf_token="mock_key")
    assert response["statusCode"] == 400
    assert response["message"] == "err.unable_to_onboard_model"



__author__ = "thiagocastroferreira"

import logging
import os
import pandas as pd
import shutil
import tarfile
import validators

from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.file_type import FileType
from aixplain.enums.storage_type import StorageType
from aixplain.modules.file import File
from aixplain.modules.metadata import MetaData
from aixplain.utils.file_utils import upload_data
from pathlib import Path
from tqdm import tqdm
from typing import List, Tuple
from urllib.parse import urlparse

AUDIO_MAX_SIZE = 50000000
IMAGE_TEXT_MAX_SIZE = 25000000


def compress_folder(folder_path: str):
    with tarfile.open(folder_path + ".tgz", "w:gz") as tar:
        for name in os.listdir(folder_path):
            tar.add(os.path.join(folder_path, name))
    return folder_path + ".tgz"


def run(metadata: MetaData, paths: List, folder: Path, batch_size: int = 100) -> Tuple[List[File], int, int, int, int]:
    """Process a list of local media files, compress and upload them to pre-signed URLs in S3

    Explanation:
        Each media on "paths" is processed. If the media is in a public link, this link is added into an index CSV file.
        If the media is in a local path, it will be copied into a local folder and its path will be added to the index CSV file.
        The medias are processed in batches such that at each "batch_size" medias, the index CSV file is uploaded into a pre-signed URL in s3 and reset.
        If the medias are stored locally, the local folder is compressed into a .tgz file and also uploaded into S3.

    Args:
        metadata (MetaData): meta data of the asset
        paths (List): list of paths to local files
        folder (Path): local folder to save compressed files before upload them to s3.

    Returns:
        Tuple[List[File], int, int, int]: list of s3 links; data, start and end columns index, and number of rows
    """
    if metadata.dtype != DataType.LABEL:
        assert (
            metadata.storage_type != StorageType.TEXT
        ), f'Data Asset Onboarding Error: Column "{metadata.name}" of type "{metadata.dtype}" can not be stored in text.'

    # if files are stored locally, create a folder to store it
    media_folder = Path(".")
    if metadata.storage_type == StorageType.FILE:
        media_folder = Path(os.path.join(folder, "data"))
        media_folder.mkdir(exist_ok=True)

    idx = 0
    data_column_idx, start_column_idx, end_column_idx = -1, -1, -1
    files, batch, start_intervals, end_intervals = [], [], [], []
    for i in tqdm(range(len(paths)), desc=f' Data "{metadata.name}" onboarding progress', position=1, leave=False):
        path = paths[i]

        if not os.path.exists(path):
            message = f'Data Asset Onboarding Error: Local file "{path}" not found.'
            logging.exception(message)
            raise Exception(message)

        dataframe = pd.read_csv(path)

        # process medias
        for j in tqdm(range(len(dataframe)), desc=" File onboarding progress", position=2, leave=False):
            row = dataframe.iloc[j]
            try:
                media_path = row[metadata.name]
            except Exception as e:
                message = f'Data Asset Onboarding Error: Column "{metadata.name}" not found in the local file "{path}".'
                logging.exception(message)
                raise Exception(message)

            # adding medias
            if metadata.storage_type == StorageType.FILE:
                if metadata.dsubtype == DataSubtype.INTERVAL:
                    # check whether the interval is in audio, image, text and video
                    assert metadata.dtype in [
                        DataType.AUDIO,
                        DataType.IMAGE,
                        DataType.TEXT,
                        DataType.VIDEO,
                    ], f'Data Asset Onboarding Error: Content Intervals do not work with "{metadata.dtype}".'
                    assert (
                        os.path.getsize(media_path) <= IMAGE_TEXT_MAX_SIZE
                    ), f'Data Asset Onboarding Error: Local interval file "{media_path}" exceeds the size limit of 25 MB.'
                    _, file_extension = os.path.splitext(media_path)
                    assert (
                        file_extension == ".json"
                    ), f'Data Asset Onboarding Error: Local interval files, such as "{media_path}", must be a JSON.'
                elif metadata.dtype == DataType.AUDIO:
                    assert (
                        os.path.getsize(media_path) <= AUDIO_MAX_SIZE
                    ), f'Data Asset Onboarding Error: Local audio file "{media_path}" exceeds the size limit of 50 MB.'
                elif metadata.dtype == DataType.LABEL:
                    assert (
                        os.path.getsize(media_path) <= IMAGE_TEXT_MAX_SIZE
                    ), f'Data Asset Onboarding Error: Local label file "{media_path}" exceeds the size limit of 25 MB.'
                else:
                    assert (
                        os.path.getsize(media_path) <= IMAGE_TEXT_MAX_SIZE
                    ), f'Data Asset Onboarding Error: Local image file "{media_path}" exceeds the size limit of 25 MB.'
                fname = os.path.basename(media_path)
                new_path = os.path.join(media_folder, fname)
                if os.path.exists(new_path) is False:
                    shutil.copy2(media_path, new_path)
                batch.append(fname)
            else:
                if metadata.storage_type == StorageType.TEXT and (
                    str(media_path).startswith("s3://")
                    or str(media_path).startswith("http://")
                    or str(media_path).startswith("https://")
                    or validators.url(media_path)
                ):
                    media_path = "DONOTDOWNLOAD" + str(media_path)
                batch.append(media_path)

            # crop intervals can not be used with interval data types
            if metadata.start_column is not None or metadata.end_column is not None:
                assert (
                    metadata.dsubtype != DataSubtype.INTERVAL
                ), f"Data Asset Onboarding Error: Interval data types can not be cropped. Remove start and end columns."

            # adding ranges to crop the media if it is the case
            if metadata.start_column is not None:
                try:
                    start_intervals.append(row[metadata.start_column])
                except Exception as e:
                    message = f'Data Asset Onboarding Error: Column "{metadata.start_column}" not found.'
                    logging.exception(message)
                    raise Exception(message)

            if metadata.end_column is not None:
                try:
                    end_intervals.append(row[metadata.end_column])
                except Exception as e:
                    message = f'Data Asset Onboarding Error: Column "{metadata.end_column}" not found.'
                    logging.exception(message)
                    raise Exception(message)

            idx += 1
            if ((idx) % batch_size) == 0:
                batch_index = str(len(files) + 1).zfill(8)

                # save index file with a list of the media files
                index_file_name = f"{folder}/{metadata.name}-{batch_index}.csv.gz"

                # if the media are stored locally, zip them and upload to s3
                if metadata.storage_type == StorageType.FILE:
                    # rename the folder where the media files are
                    data_file_name = f"{folder}/{metadata.name}-{batch_index}"
                    os.rename(media_folder, data_file_name)
                    # rename the media path in the media folder
                    for z, media_fname in enumerate(batch):
                        batch[z] = os.path.join(data_file_name, media_fname)
                    df = pd.DataFrame({metadata.name: batch})
                    # compress the folder
                    compressed_folder = compress_folder(data_file_name)
                    # upload zipped medias into s3
                    s3_compressed_folder = upload_data(compressed_folder, content_type="application/x-tar")
                    # update index files pointing the s3 link
                    df["@SOURCE"] = s3_compressed_folder
                    # remove media folder
                    shutil.rmtree(data_file_name)
                    # remove zipped file
                    os.remove(compressed_folder)
                    # create media folder again
                    media_folder = Path(os.path.join(folder, "data"))
                    media_folder.mkdir(exist_ok=True)
                else:
                    df = pd.DataFrame({metadata.name: batch})

                # adding indexes
                start, end = idx - len(batch), idx
                df["@INDEX"] = range(start, end)

                # if there are start and end time ranges, save this into the index csv
                if len(start_intervals) > 0 and len(end_intervals) > 0:
                    if metadata.dtype == DataType.AUDIO:
                        start_column = "@START_TIME"
                        end_column = "@END_TIME"
                    else:
                        start_column = "@START"
                        end_column = "@END"

                    df[start_column] = start_intervals
                    df[end_column] = end_intervals

                    start_column_idx = df.columns.to_list().index(start_column)
                    end_column_idx = df.columns.to_list().index(end_column)

                df.to_csv(index_file_name, compression="gzip", index=False)
                s3_link = upload_data(index_file_name, content_type="text/csv", content_encoding="gzip")
                files.append(File(path=s3_link, extension=FileType.CSV, compression="gzip"))
                # get data column index
                data_column_idx = df.columns.to_list().index(metadata.name)
                # restart batch variables
                batch, start_intervals, end_intervals = [], [], []

    if len(batch) > 0:
        batch_index = str(len(files) + 1).zfill(8)

        # save index file with a list of the media files
        index_file_name = f"{folder}/{metadata.name}-{batch_index}.csv.gz"

        # if the media are stored locally, zip them and upload to s3
        if metadata.storage_type == StorageType.FILE:
            # rename the folder where the media files are
            data_file_name = f"{folder}/{metadata.name}-{batch_index}"
            os.rename(media_folder, data_file_name)
            # rename the media path in the media folder
            for z, media_fname in enumerate(batch):
                batch[z] = os.path.join(data_file_name, media_fname)
            df = pd.DataFrame({metadata.name: batch})
            # compress the folder
            compressed_folder = compress_folder(data_file_name)
            # upload zipped medias into s3
            s3_compressed_folder = upload_data(compressed_folder, content_type="application/x-tar")
            # update index files pointing the s3 link
            df["@SOURCE"] = s3_compressed_folder
            # remove media folder
            shutil.rmtree(data_file_name)
            # remove zipped file
            os.remove(compressed_folder)
            # create media folder again
            media_folder = Path(os.path.join(folder, "data"))
            media_folder.mkdir(exist_ok=True)
        else:
            df = pd.DataFrame({metadata.name: batch})

        # adding indexes
        start, end = idx - len(batch), idx
        df["@INDEX"] = range(start, end)

        # if there are start and end time ranges, save this into the index csv
        if len(start_intervals) > 0 and len(end_intervals) > 0:
            if metadata.dtype == DataType.AUDIO:
                start_column = "@START_TIME"
                end_column = "@END_TIME"
            else:
                start_column = "@START"
                end_column = "@END"

            df[start_column] = start_intervals
            df[end_column] = end_intervals

            start_column_idx = df.columns.to_list().index(start_column)
            end_column_idx = df.columns.to_list().index(end_column)

        df.to_csv(index_file_name, compression="gzip", index=False)
        s3_link = upload_data(index_file_name, content_type="text/csv", content_encoding="gzip")
        files.append(File(path=s3_link, extension=FileType.CSV, compression="gzip"))
        # get data column index
        data_column_idx = df.columns.to_list().index(metadata.name)
        # restart batch variables
        batch, start_intervals, end_intervals = [], [], []
    return files, data_column_idx, start_column_idx, end_column_idx, idx



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: aiXplain team
Date: March 20th 2023
Description:
    Meta-data Class
"""

from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.file_type import FileType
from aixplain.enums.language import Language
from aixplain.enums.privacy import Privacy
from aixplain.enums.storage_type import StorageType
from typing import List, Optional, Text


class MetaData:
    def __init__(
        self,
        name: Text,
        dtype: DataType,
        storage_type: StorageType,
        data_column: Optional[Text] = None,
        start_column: Optional[Text] = None,
        end_column: Optional[Text] = None,
        privacy: Optional[Privacy] = None,
        file_extension: Optional[FileType] = None,
        languages: List[Language] = [],
        dsubtype: DataSubtype = DataSubtype.OTHER,
        id: Optional[Text] = None,
        **kwargs
    ) -> None:
        """MetaData Class

        Description:
            This class is used to stored the meta-information of the Data Class.
            It may be used to describe Data during the onboarding process of a corpus or dataset.

        Args:
            name (Text): Data Name
            dtype (DataType): Data Type
            storage_type (StorageType): Data Storage (e.g. text, local file, web link)
            data_column (Optional[Text], optional): Column index/name where the data is on a structured file (e.g. CSV). Defaults to None.
            start_column (Optional[Text], optional): Column index/name where the start indexes is on a structured file (e.g. CSV). Defaults to None.
            end_column (Optional[Text], optional): Column index/name where the end indexes is on a structured file (e.g. CSV). Defaults to None.
            privacy (Optional[Privacy], optional): Privacy of data. Defaults to None.
            file_extension (Optional[FileType], optional): File extension (e.g. CSV, TXT, etc.). Defaults to None.
            languages (List[Language], optional): List of languages which the data consists of. Defaults to [].
            dsubtype (DataSubtype, optional): Data subtype (e.g., age, topic, race, split, etc.), used in datasets metadata. Defaults to Other.
            id (Optional[Text], optional): Data ID. Defaults to None.
        """
        self.name = name
        if isinstance(dtype, str):
            dtype = DataType(dtype)
        self.dtype = dtype

        if isinstance(storage_type, str):
            storage_type = StorageType(storage_type)
        self.storage_type = storage_type

        if data_column is None:
            self.data_column = name
        else:
            self.data_column = data_column

        self.start_column = start_column
        self.end_column = end_column

        if isinstance(privacy, str):
            privacy = Privacy(privacy)
        self.privacy = privacy
        self.file_extension = file_extension

        self.languages = []
        for language in languages:
            if isinstance(language, str):
                language = Language(language)
            self.languages.append(language)
        self.dsubtype = dsubtype
        self.id = id
        self.kwargs = kwargs



__author__ = "lucaspavanelli"

"""
Copyright 2024 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Thiago Castro Ferreira and Lucas Pavanelli
Date: May 16th 2024
Description:
    Agent Factory Class
"""

import json
import logging

from aixplain.enums.function import Function
from aixplain.enums.supplier import Supplier
from aixplain.modules.agent import Agent, Tool
from aixplain.modules.agent.tool.model_tool import ModelTool
from aixplain.modules.agent.tool.pipeline_tool import PipelineTool
from aixplain.modules.model import Model
from aixplain.modules.pipeline import Pipeline
from aixplain.utils import config
from typing import Dict, List, Optional, Text, Union

from aixplain.factories.agent_factory.utils import build_agent, validate_llm
from aixplain.utils.file_utils import _request_with_retry
from urllib.parse import urljoin


class AgentFactory:
    @classmethod
    def create(
        cls,
        name: Text,
        llm_id: Text,
        tools: List[Tool] = [],
        description: Text = "",
        api_key: Text = config.TEAM_API_KEY,
        supplier: Union[Dict, Text, Supplier, int] = "aiXplain",
        version: Optional[Text] = None,
        use_mentalist_and_inspector: bool = False,
    ) -> Agent:
        """Create a new agent in the platform.

        Args:
            name (Text): name of the agent
            llm_id (Text): aiXplain ID of the large language model to be used as agent.
            tools (List[Tool], optional): list of tool for the agent. Defaults to [].
            description (Text, optional): description of the agent role. Defaults to "".
            api_key (Text, optional): team/user API key. Defaults to config.TEAM_API_KEY.
            supplier (Union[Dict, Text, Supplier, int], optional): owner of the agent. Defaults to "aiXplain".
            version (Optional[Text], optional): version of the agent. Defaults to None.
            use_mentalist_and_inspector (bool, optional): flag to enable mentalist and inspector agents (which only works when a supervisor is enabled). Defaults to False.

        Returns:
            Agent: created Agent
        """
        # validate LLM ID
        validate_llm(llm_id)

        orchestrator_llm_id, mentalist_and_inspector_llm_id = llm_id, None
        if use_mentalist_and_inspector is True:
            mentalist_and_inspector_llm_id = llm_id

        try:
            agent = None
            url = urljoin(config.BACKEND_URL, "sdk/agents")
            headers = {"x-api-key": api_key}

            if isinstance(supplier, dict):
                supplier = supplier["code"]
            elif isinstance(supplier, Supplier):
                supplier = supplier.value["code"]

            tool_payload = []
            for tool in tools:
                if isinstance(tool, ModelTool):
                    tool.validate()
                    tool_payload.append(
                        {
                            "function": tool.function.value if tool.function is not None else None,
                            "type": "model",
                            "description": tool.description,
                            "supplier": tool.supplier.value["code"] if tool.supplier else None,
                            "version": tool.version if tool.version else None,
                            "assetId": tool.model,
                        }
                    )
                elif isinstance(tool, PipelineTool):
                    tool.validate()
                    tool_payload.append(
                        {
                            "assetId": tool.pipeline,
                            "description": tool.description,
                            "type": "pipeline",
                        }
                    )
                else:
                    raise Exception("Agent Creation Error: Tool type not supported.")

            payload = {
                "name": name,
                "assets": tool_payload,
                "description": description,
                "supplier": supplier,
                "version": version,
                "llmId": llm_id,
                "supervisorId": orchestrator_llm_id,
                "plannerId": mentalist_and_inspector_llm_id,
            }

            logging.info(f"Start service for POST Create Agent  - {url} - {headers} - {json.dumps(payload)}")
            r = _request_with_retry("post", url, headers=headers, json=payload)
            if 200 <= r.status_code < 300:
                response = r.json()
                agent = build_agent(payload=response, api_key=api_key)
            else:
                error = r.json()
                error_msg = "Agent Onboarding Error: Please contant the administrators."
                if "message" in error:
                    msg = error["message"]
                    if error["message"] == "err.name_already_exists":
                        msg = "Agent name already exists."
                    elif error["message"] == "err.asset_is_not_available":
                        msg = "Some the tools are not available."
                    error_msg = f"Agent Onboarding Error (HTTP {r.status_code}): {msg}"
                logging.exception(error_msg)
                raise Exception(error_msg)
        except Exception as e:
            raise Exception(e)
        return agent

    @classmethod
    def create_model_tool(
        cls,
        model: Optional[Union[Model, Text]] = None,
        function: Optional[Union[Function, Text]] = None,
        supplier: Optional[Union[Supplier, Text]] = None,
    ) -> ModelTool:
        """Create a new model tool."""
        if function is not None and isinstance(function, str):
            function = Function(function)

        if supplier is not None:
            if isinstance(supplier, str):
                for supplier_ in Supplier:
                    if supplier.lower() in [supplier.value["code"].lower(), supplier.value["name"].lower()]:
                        supplier = supplier_
                        break
                if isinstance(supplier, str):
                    supplier = None
        return ModelTool(function=function, supplier=supplier, model=model)

    @classmethod
    def create_pipeline_tool(cls, description: Text, pipeline: Union[Pipeline, Text]) -> PipelineTool:
        """Create a new pipeline tool."""
        return PipelineTool(description=description, pipeline=pipeline)

    @classmethod
    def list(cls) -> Dict:
        """List all agents available in the platform."""
        url = urljoin(config.BACKEND_URL, "sdk/agents")
        headers = {"x-api-key": config.TEAM_API_KEY, "Content-Type": "application/json"}

        payload = {}
        logging.info(f"Start service for GET List Agents - {url} - {headers} - {json.dumps(payload)}")
        try:
            r = _request_with_retry("get", url, headers=headers)
            resp = r.json()

            if 200 <= r.status_code < 300:
                agents, page_total, total = [], 0, 0
                results = resp
                page_total = len(results)
                total = len(results)
                logging.info(f"Response for GET List Agents - Page Total: {page_total} / Total: {total}")
                for agent in results:
                    agents.append(build_agent(agent))
                return {"results": agents, "page_total": page_total, "page_number": 0, "total": total}
            else:
                error_msg = "Agent Listing Error: Please contant the administrators."
                if "message" in resp:
                    msg = resp["message"]
                    error_msg = f"Agent Listing Error (HTTP {r.status_code}): {msg}"
                logging.exception(error_msg)
                raise Exception(error_msg)
        except Exception as e:
            raise Exception(e)

    @classmethod
    def get(cls, agent_id: Text, api_key: Optional[Text] = None) -> Agent:
        """Get agent by id."""
        url = urljoin(config.BACKEND_URL, f"sdk/agents/{agent_id}")
        if config.AIXPLAIN_API_KEY != "":
            headers = {"x-aixplain-key": f"{config.AIXPLAIN_API_KEY}", "Content-Type": "application/json"}
        else:
            api_key = api_key if api_key is not None else config.TEAM_API_KEY
            headers = {"x-api-key": api_key, "Content-Type": "application/json"}
        logging.info(f"Start service for GET Agent  - {url} - {headers}")
        r = _request_with_retry("get", url, headers=headers)
        resp = r.json()
        if 200 <= r.status_code < 300:
            return build_agent(resp)
        else:
            msg = "Please contant the administrators."
            if "message" in resp:
                msg = resp["message"]
            error_msg = f"Agent Get Error (HTTP {r.status_code}): {msg}"
            raise Exception(error_msg)



from aixplain.utils import config


def check_api_key(method):
    def wrapper(*args, **kwargs):
        if config.TEAM_API_KEY == "":
            raise Exception(
                "A 'TEAM_API_KEY' is required to run an asset. For help, please refer to the documentation (https://github.com/aixplain/aixplain#api-key-setup)"
            )
        return method(*args, **kwargs)

    return wrapper




from dotenv import load_dotenv
from urllib.parse import urljoin
import requests_mock
from aixplain.enums import Function

load_dotenv()
from aixplain.utils import config
from aixplain.modules import LLM

import pytest

@pytest.mark.parametrize(
    "status_code,error_message",
    [
        (401,"Unauthorized API key: Please verify the spelling of the API key and its current validity."),
        (465,"Subscription-related error: Please ensure that your subscription is active and has not expired."),
        (475,"Billing-related error: Please ensure you have enough credits to run this model. "),
        (485, "Supplier-related error: Please ensure that the selected supplier provides the model you are trying to access."),
        (495, "Validation-related error: Please ensure all required fields are provided and correctly formatted."),
        (501, "Status 501: Unspecified error: An unspecified error occurred while processing your request."),

    ],
)

def test_run_async_errors(status_code, error_message):
    base_url = config.MODELS_RUN_URL
    llm_id = "llm-id"
    execute_url = urljoin(base_url, f"execute/{llm_id}")
    
    with requests_mock.Mocker() as mock:
        mock.post(execute_url, status_code=status_code)
        test_llm = LLM(id=llm_id, name="Test llm",url=base_url, function=Function.TEXT_GENERATION)
        response = test_llm.run_async(data="input_data")
    assert response["status"] == "FAILED"
    assert response["error_message"] == error_message


__author__ = "lucaspavanelli"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: June 14th 2023
Description:
    FineTune Class
"""
from typing import List, Text, Optional
import logging
import json
from urllib.parse import urljoin
from aixplain.modules.finetune.cost import FinetuneCost
from aixplain.modules.finetune.hyperparameters import Hyperparameters
from aixplain.modules.asset import Asset
from aixplain.modules.dataset import Dataset
from aixplain.modules.model import Model

from aixplain.utils import config
from aixplain.utils.file_utils import _request_with_retry


class Finetune(Asset):
    """FineTune is a powerful tool for fine-tuning machine learning models and using your own datasets for specific tasks.

    Attributes:
        name (Text): Name of the FineTune.
        dataset_list (List[Dataset]): List of Datasets to be used for fine-tuning.
        model (Model): Model to be fine-tuned.
        cost (Cost): Cost of the FineTune.
        id (Text): ID of the FineTune.
        description (Text): Description of the FineTune.
        supplier (Text): Supplier of the FineTune.
        version (Text): Version of the FineTune.
        train_percentage (float): Percentage of training samples.
        dev_percentage (float): Percentage of development samples.
        prompt_template (Text): Fine-tuning prompt_template.
        hyperparameters (Hyperparameters): Hyperparameters for fine-tuning.
        additional_info (dict): Additional information to be saved with the FineTune.
        backend_url (str): URL of the backend.
        api_key (str): The TEAM API key used for authentication.
    """

    def __init__(
        self,
        name: Text,
        dataset_list: List[Dataset],
        model: Model,
        cost: FinetuneCost,
        id: Optional[Text] = "",
        description: Optional[Text] = "",
        supplier: Optional[Text] = "aiXplain",
        version: Optional[Text] = "1.0",
        train_percentage: Optional[float] = 100,
        dev_percentage: Optional[float] = 0,
        prompt_template: Optional[Text] = None,
        hyperparameters: Optional[Hyperparameters] = None,
        **additional_info,
    ) -> None:
        """Create a FineTune with the necessary information.

        Args:
            name (Text): Name of the FineTune.
            dataset_list (List[Dataset]): List of Datasets to be used for fine-tuning.
            model (Model): Model to be fine-tuned.
            cost (Cost): Cost of the FineTune.
            id (Text, optional): ID of the FineTune. Defaults to "".
            description (Text, optional): Description of the FineTune. Defaults to "".
            supplier (Text, optional): Supplier of the FineTune. Defaults to "aiXplain".
            version (Text, optional): Version of the FineTune. Defaults to "1.0".
            train_percentage (float, optional): Percentage of training samples. Defaults to 100.
            dev_percentage (float, optional): Percentage of development samples. Defaults to 0.
            prompt_template (Text, optional): Fine-tuning prompt_template. Should reference columns in the dataset using format <<COLUMN_NAME>>. Defaults to None.
            hyperparameters (Hyperparameters, optional): Hyperparameters for fine-tuning. Defaults to None.
            **additional_info: Additional information to be saved with the FineTune.
        """
        super().__init__(id, name, description, supplier, version)
        self.model = model
        self.dataset_list = dataset_list
        self.cost = cost
        self.train_percentage = train_percentage
        self.dev_percentage = dev_percentage
        self.prompt_template = prompt_template
        self.hyperparameters = hyperparameters
        self.additional_info = additional_info
        self.backend_url = config.BACKEND_URL
        self.api_key = config.TEAM_API_KEY
        self.aixplain_key = config.AIXPLAIN_API_KEY

    def start(self) -> Model:
        """Start the Finetune job.

        Returns:
            Model: The model object representing the Finetune job.
        """
        payload = {}
        try:
            url = urljoin(self.backend_url, "sdk/finetune")
            headers = {"Authorization": f"Token {self.api_key}", "Content-Type": "application/json"}
            payload = {
                "name": self.name,
                "datasets": [
                    {
                        "datasetId": dataset.id,
                        "trainSamplesPercentage": self.train_percentage,
                        "devSamplesPercentage": self.dev_percentage,
                    }
                    for dataset in self.dataset_list
                ],
                "sourceModelId": self.model.id,
            }
            parameters = {}
            if self.prompt_template is not None:
                parameters["prompt"] = self.prompt_template
            if self.hyperparameters is not None:
                parameters["hyperparameters"] = self.hyperparameters.to_dict()
            payload["parameters"] = parameters
            logging.info(f"Start service for POST Start FineTune - {url} - {headers} - {json.dumps(payload)}")
            r = _request_with_retry("post", url, headers=headers, json=payload)
            resp = r.json()
            logging.info(f"Response for POST Start FineTune - Name: {self.name} / Status {resp}")
            from aixplain.factories.model_factory import ModelFactory

            return ModelFactory().get(resp["id"])
        except Exception:
            message = ""
            if resp is not None and "statusCode" in resp:
                status_code = resp["statusCode"]
                message = resp["message"]
                message = f"Status {status_code} - {message}"
            error_message = f"Start FineTune: Error with payload {json.dumps(payload)}: {message}"
            logging.exception(error_message)
            return None



from typing import (
    List,
    Union,
    TYPE_CHECKING,
    Generic,
    TypeVar,
    Type,
    Optional,
    Iterator,
)

from aixplain.enums import DataType
from .enums import NodeType, ParamType


if TYPE_CHECKING:
    from .pipeline import DesignerPipeline

TI = TypeVar("TI", bound="Inputs")
TO = TypeVar("TO", bound="Outputs")


class Serializable:
    def serialize(self) -> dict:
        raise NotImplementedError()


class Param(Serializable):
    """
    Param class, this class will be used to create the parameters of the node.
    """

    code: str
    param_type: ParamType
    data_type: Optional[DataType] = None
    value: Optional[str] = None
    node: Optional["Node"] = None
    link_: Optional["Link"] = None

    def __init__(
        self,
        code: str,
        data_type: Optional[DataType] = None,
        value: Optional[str] = None,
        node: Optional["Node"] = None,
        param_type: Optional[ParamType] = None,
    ):
        self.code = code
        self.data_type = data_type
        self.value = value

        # is subclasses do not set the param type, set it to None
        self.param_type = getattr(self, "param_type", param_type)

        if node:
            self.attach_to(node)

    def attach_to(self, node: "Node") -> "Param":
        """
        Attach the param to the node.
        :param node: the node
        :return: the param
        """
        assert not self.node, "Param already attached to a node"
        assert self.param_type, "Param type not set"
        if self.param_type == ParamType.INPUT:
            node.inputs.add_param(self)
        elif self.param_type == ParamType.OUTPUT:
            node.outputs.add_param(self)
        else:
            raise ValueError("Invalid param type")
        self.node = node
        return self

    def link(self, to_param: "Param") -> "Param":
        """
        Link the output of the param to the input of another param.
        :param to_param: the input param
        :return: the param
        """
        assert self.node, "Param not attached to a node"
        assert to_param.param_type == ParamType.INPUT, "Invalid param type"
        assert self in self.node.outputs, "Param not registered as output"
        return to_param.back_link(self)

    def back_link(self, from_param: "Param") -> "Param":
        """
        Link the input of the param to the output of another param.
        :param from_param: the output param
        :return: the param
        """
        assert self.node, "Param not attached to a node"
        assert from_param.param_type == ParamType.OUTPUT, "Invalid param type"
        assert self.code in self.node.inputs, "Param not registered as input"
        link = from_param.node.link(self.node, from_param, self)
        self.link_ = link
        from_param.link_ = link
        return link

    def serialize(self) -> dict:
        return {
            "code": self.code,
            "dataType": self.data_type,
            "value": self.value,
        }


class InputParam(Param):

    param_type: ParamType = ParamType.INPUT
    is_required: bool = True

    def __init__(self, *args, is_required: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_required = is_required


class OutputParam(Param):

    param_type: ParamType = ParamType.OUTPUT


class Link(Serializable):
    """
    Link class, this class will be used to link the output of the node to the
    input of another node.
    """

    from_node: "Node"
    to_node: "Node"
    from_param: str
    to_param: str

    pipeline: Optional["DesignerPipeline"] = None

    def __init__(
        self,
        from_node: "Node",
        to_node: "Node",
        from_param: Union[Param, str],
        to_param: Union[Param, str],
        pipeline: "DesignerPipeline" = None,
    ):

        assert from_param in from_node.outputs, "Invalid from param"
        assert to_param in to_node.inputs, "Invalid to param"

        if isinstance(from_param, Param):
            from_param = from_param.code
        if isinstance(to_param, Param):
            to_param = to_param.code

        self.from_node = from_node
        self.to_node = to_node
        self.from_param = from_param
        self.to_param = to_param

        if pipeline:
            self.attach_to(pipeline)

        # self.validate()
        self.auto_infer()

    def auto_infer(self):
        from_param = self.from_node.outputs[self.from_param]
        to_param = self.to_node.inputs[self.to_param]

        # if one of the data types is missing, infer the other one
        data_type = from_param.data_type or to_param.data_type
        from_param.data_type = data_type
        to_param.data_type = data_type

        def infer_data_type(node):
            from .nodes import Input, Output

            if isinstance(node, Input) or isinstance(node, Output):
                if data_type and data_type not in node.data_types:
                    node.data_types.append(data_type)

        infer_data_type(self.from_node)
        infer_data_type(self.to_node)

    def validate(self):
        from_param = self.from_node.outputs[self.from_param]
        to_param = self.to_node.inputs[self.to_param]

        # Should we check for data type mismatch?
        if from_param.data_type and to_param.data_type:
            if from_param.data_type != to_param.data_type:
                raise ValueError(
                    f"Data type mismatch between {from_param.data_type} and {to_param.data_type}"
                )  # noqa

    def attach_to(self, pipeline: "DesignerPipeline"):
        """
        Attach the link to the pipeline.
        :param pipeline: the pipeline
        """
        assert not self.pipeline, "Link already attached to a pipeline"
        if not self.from_node.pipeline or self.from_node not in pipeline.nodes:
            self.from_node.attach_to(pipeline)
        if not self.to_node.pipeline or self.to_node not in pipeline.nodes:
            self.to_node.attach_to(pipeline)

        self.pipeline = pipeline
        self.pipeline.links.append(self)
        return self

    def serialize(self) -> dict:
        assert self.from_node.number is not None, "From node number not set"
        assert self.to_node.number is not None, "To node number not set"
        return {
            "from": self.from_node.number,
            "to": self.to_node.number,
            "paramMapping": [
                {
                    "from": self.from_param,
                    "to": self.to_param,
                }
            ],
        }


class ParamProxy(Serializable):

    node: "Node"

    def __init__(self, node: "Node", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.node = node
        self._params = []

    def add_param(self, param: Param) -> None:
        # check if param already registered
        if param in self:
            raise ValueError(
                f"Parameter with code '{param.code}' already exists."
            )
        self._params.append(param)
        # also set attribute on the node dynamically if there's no
        # any attribute with the same name
        if not hasattr(self, param.code):
            setattr(self, param.code, param)

    def _create_param(
        self, code: str, data_type: DataType = None, value: any = None
    ) -> Param:
        raise NotImplementedError()

    def create_param(
        self,
        code: str,
        data_type: DataType = None,
        value: any = None,
        is_required: bool = False,
    ) -> Param:
        param = self._create_param(code, data_type, value)
        param.is_required = is_required
        self.add_param(param)
        param.node = self.node
        return param

    def __getitem__(self, code: str) -> Param:
        for param in self._params:
            if param.code == code:
                return param
        raise KeyError(f"Parameter with code '{code}' not found.")

    def __setitem__(self, code: str, value: str) -> None:
        # set param value on set item to avoid setting it manually
        self[code].value = value

    def __setattr__(self, name: str, value: any) -> None:
        # set param value on attribute assignment to avoid setting it manually
        if isinstance(value, str) and hasattr(self, name):
            self[name].value = value
        else:
            super().__setattr__(name, value)

    def __contains__(self, param: Union[str, Param]) -> bool:
        code = param if isinstance(param, str) else param.code
        return any(param.code == code for param in self._params)

    def __iter__(self) -> Iterator[Param]:
        return iter(self._params)

    def __len__(self) -> int:
        return len(self._params)

    def serialize(self) -> List[dict]:
        return [param.serialize() for param in self._params]


class Inputs(ParamProxy):
    def _create_param(
        self,
        code: str,
        data_type: DataType = None,
        value: any = None,
        is_required: bool = False,
    ) -> InputParam:
        return InputParam(
            code=code,
            data_type=data_type,
            value=value,
            is_required=is_required,
        )


class Outputs(ParamProxy):
    def _create_param(
        self, code: str, data_type: DataType = None, value: any = None
    ) -> OutputParam:
        return OutputParam(code=code, data_type=data_type, value=value)


class Node(Generic[TI, TO], Serializable):
    """
    Node class is the base class for all the nodes in the pipeline. This class
    will be used to create the nodes and link them together.
    """

    number: Optional[int] = None
    label: Optional[str] = None
    type: Optional[NodeType] = None

    inputs: Optional[TI] = None
    outputs: Optional[TO] = None
    inputs_class: Optional[Type[TI]] = Inputs
    outputs_class: Optional[Type[TO]] = Outputs
    pipeline: Optional["DesignerPipeline"] = None

    def __init__(
        self,
        pipeline: "DesignerPipeline" = None,
        number: Optional[int] = None,
        label: Optional[str] = None,
    ):
        self.inputs = self.inputs_class(node=self)
        self.outputs = self.outputs_class(node=self)
        self.number = number
        self.label = label

        if pipeline:
            self.attach_to(pipeline)

    def attach_to(self, pipeline: "DesignerPipeline"):
        """
        Attach the node to the pipeline.
        :param pipeline: the pipeline
        """
        assert not self.pipeline, "Node already attached to a pipeline"
        assert (
            self not in pipeline.nodes
        ), "Node already attached to a pipeline"
        assert self.type, "Node type not set"

        self.pipeline = pipeline
        if self.number is None:
            self.number = len(pipeline.nodes)
        if self.label is None:
            self.label = f"{self.type.value}(ID={self.number})"

        assert not pipeline.get_node(self.number), "Node number already exists"
        pipeline.nodes.append(self)
        return self

    def serialize(self) -> dict:
        return {
            "number": self.number,
            "label": self.label,
            "type": self.type.value,
            "inputValues": self.inputs.serialize(),
            "outputValues": self.outputs.serialize(),
        }



__author__ = "shreyassharma"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: December 1st 2022
Description:
    Dataset Factory Class
"""

import aixplain.utils.config as config
import aixplain.processes.data_onboarding.onboard_functions as onboard_functions
import json
import os
import logging
import shutil

from aixplain.factories.asset_factory import AssetFactory
from aixplain.factories.data_factory import DataFactory
from aixplain.modules.data import Data
from aixplain.modules.dataset import Dataset
from aixplain.modules.metadata import MetaData
from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.error_handler import ErrorHandler
from aixplain.enums.function import Function
from aixplain.enums.language import Language
from aixplain.enums.license import License
from aixplain.enums.privacy import Privacy
from aixplain.utils import config
from aixplain.utils.convert_datatype_utils import dict_to_metadata
from aixplain.utils.file_utils import _request_with_retry, s3_to_csv
from aixplain.utils.validation_utils import dataset_onboarding_validation
from pathlib import Path
from tqdm import tqdm
from typing import Any, Dict, List, Optional, Text, Union
from urllib.parse import urljoin
from uuid import uuid4
from warnings import warn


class DatasetFactory(AssetFactory):
    """A static class for creating and exploring Dataset Objects.

    Attributes:
        backend_url (str): The URL for the backend.
    """

    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def __from_response(cls, response: Dict) -> Dataset:
        """Converts response Json to 'Dataset' object

        Args:
            response (dict): Json from API

        Returns:
            Dataset: Coverted 'Dataset' object
        """
        # process data
        data = {}
        for d in response["data"]:
            languages = []
            if "languages" in d["metadata"]:
                languages = []
                for lng in d["metadata"]["languages"]:
                    if "dialect" not in lng:
                        lng["dialect"] = ""
                    languages.append(Language(lng))

            data[d["id"]] = Data(
                id=d["id"],
                name=d["name"],
                dtype=DataType(d["dataType"]),
                dsubtype=DataSubtype(d["dataSubtype"]),
                privacy=Privacy.PRIVATE,
                languages=languages,
                onboard_status=d["status"],
            )

        # process input data
        source_data = {}
        for inp in response["input"]:
            data_id = inp["dataId"]
            if data_id in data:
                source_data[data[data_id].name] = data[data_id]

        # process hypotheses
        hypotheses = {}
        if "hypotheses" in response:
            for inp in response["hypotheses"]:
                data_id = inp["dataId"]
                if data_id in data:
                    hypotheses[data[data_id].name] = data[data_id]

        # process metadata
        metadata = {}
        for inp in response["metadata"]:
            data_id = inp["dataId"]
            if data_id in data:
                metadata[data[data_id].name] = data[data_id]

        # process output data
        target_data = {}
        for out in response["output"]:
            try:
                target_data_list = [data[data_id] for data_id in out["dataIds"]]
                data_name = target_data_list[0].name
                target_data[data_name] = target_data_list
            except:
                pass

        # process function
        function = Function(response["function"])

        # process license
        try:
            license = License(response["license"]["typeId"])
        except Exception:
            license = None

        try:
            length = int(response["segmentsCount"])
        except Exception:
            length = None

        dataset = Dataset(
            id=response["id"],
            name=response["name"],
            description=response["description"],
            function=function,
            license=license,
            source_data=source_data,
            target_data=target_data,
            hypotheses=hypotheses,
            metadata=metadata,
            onboard_status=response["status"],
            length=length,
        )
        return dataset

    @classmethod
    def get(cls, dataset_id: Text) -> Dataset:
        """Create a 'Dataset' object from dataset id

        Args:
            dataset_id (Text): Dataset ID of required dataset.

        Returns:
            Dataset: Created 'Dataset' object
        """
        url = urljoin(cls.backend_url, f"sdk/datasets/{dataset_id}/overview")
        if cls.aixplain_key != "":
            headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        logging.info(f"Start service for GET Dataset  - {url} - {headers}")
        r = _request_with_retry("get", url, headers=headers)
        resp = r.json()
        if "statusCode" in resp and resp["statusCode"] == 404:
            raise Exception(f"Dataset GET Error: Dataset {dataset_id} not found.")
        return cls.__from_response(resp)

    @classmethod
    def list(
        cls,
        query: Optional[Text] = None,
        function: Optional[Function] = None,
        source_languages: Optional[Union[Language, List[Language]]] = None,
        target_languages: Optional[Union[Language, List[Language]]] = None,
        data_type: Optional[DataType] = None,
        license: Optional[License] = None,
        is_referenceless: Optional[bool] = None,
        page_number: int = 0,
        page_size: int = 20,
    ) -> Dict:
        """Listing Datasets

        Args:
            query (Optional[Text], optional): search query. Defaults to None.
            function (Optional[Function], optional): function filter. Defaults to None.
            source_languages (Optional[Union[Language, List[Language]]], optional): language filter of input data. Defaults to None.
            target_languages (Optional[Union[Language, List[Language]]], optional): language filter of output data. Defaults to None.
            data_type (Optional[DataType], optional): data type filter. Defaults to None.
            license (Optional[License], optional): license filter. Defaults to None.
            is_referenceless (Optional[bool], optional): has reference filter. Defaults to None.
            page_number (int, optional): page number. Defaults to 0.
            page_size (int, optional): page size. Defaults to 20.

        Returns:
            Dict: list of datasets in agreement with the filters, page number, page total and total elements
        """
        url = urljoin(cls.backend_url, "sdk/datasets/paginate")
        if cls.aixplain_key != "":
            headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}

        assert 0 < page_size <= 100, f"Dataset List Error: Page size must be greater than 0 and not exceed 100."
        payload = {
            "pageSize": page_size,
            "pageNumber": page_number,
            "sort": [{"field": "createdAt", "dir": -1}],
            "input": {},
            "output": {},
        }

        if query is not None:
            payload["q"] = str(query)

        if function is not None:
            payload["function"] = function.value

        if license is not None:
            payload["license"] = license.value

        if data_type is not None:
            payload["dataType"] = data_type.value

        if is_referenceless is not None:
            payload["isReferenceless"] = is_referenceless

        if source_languages is not None:
            if isinstance(source_languages, Language):
                source_languages = [source_languages]
            payload["input"]["languages"] = [lng.value["language"] for lng in source_languages]

        if target_languages is not None:
            if isinstance(target_languages, Language):
                target_languages = [target_languages]
            payload["output"]["languages"] = [lng.value["language"] for lng in target_languages]

        logging.info(f"Start service for POST List Dataset - {url} - {headers} - {json.dumps(payload)}")
        r = _request_with_retry("post", url, headers=headers, json=payload)
        resp = r.json()

        datasets, page_total, total = [], 0, 0
        if "results" in resp:
            results = resp["results"]
            page_total = resp["pageTotal"]
            total = resp["total"]
            logging.info(f"Response for POST List Dataset - Page Total: {page_total} / Total: {total}")
            for dataset in results:
                datasets.append(cls.__from_response(dataset))
        return {"results": datasets, "page_total": page_total, "page_number": page_number, "total": total}

    @classmethod
    def create(
        cls,
        name: Text,
        description: Text,
        license: License,
        function: Function,
        input_schema: List[Union[Dict, MetaData]],
        output_schema: List[Union[Dict, MetaData]] = [],
        hypotheses_schema: List[Union[Dict, MetaData]] = [],
        metadata_schema: List[Union[Dict, MetaData]] = [],
        content_path: Union[Union[Text, Path], List[Union[Text, Path]]] = [],
        input_ref_data: Dict[Text, Any] = {},
        output_ref_data: Dict[Text, List[Any]] = {},
        hypotheses_ref_data: Dict[Text, Any] = {},
        meta_ref_data: Dict[Text, Any] = {},
        tags: List[Text] = [],
        privacy: Privacy = Privacy.PRIVATE,
        split_labels: Optional[List[Text]] = None,
        split_rate: Optional[List[float]] = None,
        error_handler: ErrorHandler = ErrorHandler.SKIP,
        s3_link: Optional[Text] = None,
        aws_credentials: Optional[Dict[Text, Text]] = {"AWS_ACCESS_KEY_ID": None, "AWS_SECRET_ACCESS_KEY": None},
        api_key: Optional[Text] = None
    ) -> Dict:
        """Dataset Onboard

        Args:
            name (Text): dataset name
            description (Text): dataset description
            license (License): dataset license
            function (Function): dataset function
            input_schema (List[Union[Dict, MetaData]]): metadata of inputs
            output_schema (List[Union[Dict, MetaData]]): metadata of outputs
            hypotheses_schema (List[Union[Dict, MetaData]], optional): schema of the hypotheses to the references. Defaults to [].
            metadata_schema (List[Union[Dict, MetaData]], optional): metadata of metadata information of the dataset. Defaults to [].
            content_path (Union[Union[Text, Path], List[Union[Text, Path]]]): path to files which contain the data content
            input_ref_data (Dict[Text, Any], optional): reference to input data which is already in the platform. Defaults to {}.
            output_ref_data (Dict[Text, List[Any]], optional): reference to output data which is already in the platform. Defaults to {}.
            hypotheses_ref_data (Dict[Text, Any], optional): hypotheses which are already in the platform. Defaults to {}.
            meta_ref_data (Dict[Text, Any], optional): metadata which is already in the platform. Defaults to {}.
            tags (List[Text], optional): datasets description tags. Defaults to [].
            privacy (Privacy, optional): dataset privacy. Defaults to Privacy.PRIVATE.
            error_handler (ErrorHandler, optional): how to handle failed rows in the data asset. Defaults to ErrorHandler.SKIP.
            s3_link (Optional[Text]): s3 url to files or directories
            aws_credentials (Optional[Dict[Text, Text]]) : credentials for AWS and it should contains these two keys `AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY`
            api_key (Optional[Text]): team api key. Defaults to None.
        Returns:
            Dict: dataset onboard status
        """

        for lmd in (hypotheses_schema, input_schema, output_schema, metadata_schema):
            dict_to_metadata(lmd)

        dataset_onboarding_validation(
            input_schema=input_schema,
            output_schema=output_schema,
            function=function,
            input_ref_data=input_ref_data,
            metadata_schema=metadata_schema,
            content_path=content_path,
            split_labels=split_labels,
            split_rate=split_rate,
            s3_link=s3_link,
        )

        folder, return_dict, ref_data, csv_path = None, {}, [], None
        # check team key
        try:
            # process data and create files
            folder = Path(name)
            folder.mkdir(exist_ok=True)

            if isinstance(content_path, list) is False:
                content_paths = [content_path]
            else:
                content_paths = content_path

            if s3_link is not None:
                csv_path = s3_to_csv(s3_link, aws_credentials)
                content_paths.append(csv_path)

            # assert (
            #     len(output_schema) > 0 or len(output_ref_data) > 0
            # ), "Data Asset Onboarding Error: You must specify an output data to onboard a dataset."

            for input_data in input_ref_data:
                if isinstance(input_ref_data[input_data], Data):
                    input_ref_data[input_data] = input_ref_data[input_data].id
                # check whether the referred data exist. Otherwise, raise an exception
                data_id = input_ref_data[input_data]
                if onboard_functions.is_data(data_id) is False:
                    message = f"Data Asset Onboarding Error: Referenced Input Data {data_id} does not exist."
                    logging.exception(message)
                    raise Exception(message)
                ref_data.append(DataFactory.get(data_id=data_id))

            for output_data in output_ref_data:
                if len(output_ref_data[output_data]) > 0:
                    if isinstance(output_ref_data[output_data][0], Data):
                        output_ref_data[output_data] = [w.id for w in output_ref_data[output_data]]
                    # check whether the referred data exist. Otherwise, raise an exception
                    for data_id in output_ref_data[output_data]:
                        if onboard_functions.is_data(data_id) is False:
                            message = f"Data Asset Onboarding Error: Referenced Output Data {data_id} does not exist."
                            logging.exception(message)
                            raise Exception(message)
                        ref_data.append(DataFactory.get(data_id=data_id))

            for hypdata in hypotheses_ref_data:
                if isinstance(hypotheses_ref_data[hypdata], Data):
                    hypotheses_ref_data[hypdata] = hypotheses_ref_data[hypdata].id
                # check whether the referred data exist. Otherwise, raise an exception
                data_id = hypotheses_ref_data[hypdata]
                if onboard_functions.is_data(data_id) is False:
                    message = f"Data Asset Onboarding Error: Referenced Hypotheses Data {data_id} does not exist."
                    logging.exception(message)
                    raise Exception(message)
                ref_data.append(DataFactory.get(data_id=data_id))

            for meta_data in meta_ref_data:
                if isinstance(meta_ref_data[meta_data], Data):
                    meta_ref_data[meta_data] = meta_ref_data[meta_data].id
                # check whether the referred data exist. Otherwise, raise an exception
                data_id = meta_ref_data[meta_data]
                if onboard_functions.is_data(data_id) is False:
                    message = f"Data Asset Onboarding Error: Referenced Meta Data {data_id} does not exist."
                    logging.exception(message)
                    raise Exception(message)
                ref_data.append(DataFactory.get(data_id=data_id))

            # check whether reserved names are used as data/column names
            for schema in [input_schema, output_schema, metadata_schema, hypotheses_schema]:
                for metadata in schema:
                    for forbidden_name in onboard_functions.FORBIDDEN_COLUMN_NAMES:
                        if forbidden_name in [metadata.name, metadata.data_column]:
                            message = f"Data Asset Onboarding Error: {forbidden_name} is reserved name and must not be used as the name of a data or a column."
                            logging.exception(message)
                            raise Exception(message)

            # get file extension paths to process
            paths = onboard_functions.get_paths(content_paths)

            # set dataset split
            if split_labels is not None and split_rate is not None:
                assert len(split_labels) == len(
                    split_rate
                ), "Data Asset Onboarding Error: Make sure you set the *split_labels* and *split_rate* lists must have the same length."
                split_metadata = onboard_functions.split_data(paths=paths, split_labels=split_labels, split_rate=split_rate)
                metadata_schema.append(split_metadata)

            datasets, sizes = {}, []
            for (key, schema) in [
                ("inputs", input_schema),
                ("outputs", output_schema),
                ("hypotheses", hypotheses_schema),
                ("meta", metadata_schema),
            ]:
                datasets[key] = {}
                for i in tqdm(range(len(schema)), desc=f" Dataset's {key} onboard progress", position=0):
                    metadata = schema[i]
                    if metadata.privacy is None:
                        metadata.privacy = privacy

                    files, data_column_idx, start_column_idx, end_column_idx, nrows = onboard_functions.process_data_files(
                        data_asset_name=name, metadata=metadata, paths=paths, folder=name
                    )

                    # save size
                    sizes.append(nrows)

                    if metadata.name not in datasets[key]:
                        datasets[key][metadata.name] = []

                    datasets[key][metadata.name].append(
                        Data(
                            id=str(uuid4()).replace("-", ""),
                            name=metadata.name,
                            dtype=metadata.dtype,
                            dsubtype=metadata.dsubtype,
                            privacy=metadata.privacy,
                            onboard_status="onboarding",
                            data_column=data_column_idx,
                            start_column=start_column_idx,
                            end_column=end_column_idx,
                            files=files,
                            languages=metadata.languages,
                            length=nrows,
                        )
                    )

            # validate and flat inputs, hypotheses and metadata
            for key_schema in ["inputs", "hypotheses", "meta"]:
                for input_data in datasets[key_schema]:
                    assert len(datasets[key_schema][input_data]) == 1
                    datasets[key_schema][input_data] = datasets[key_schema][input_data][0]

            dataset = Dataset(
                id="",
                name=name,
                description=description,
                function=function,
                source_data=datasets["inputs"],
                target_data=datasets["outputs"],
                hypotheses=datasets["hypotheses"],
                metadata=datasets["meta"],
                tags=tags,
                license=license,
                privacy=privacy,
                onboard_status="onboarding",
            )

            # check alignment
            sizes += [d.length for d in ref_data]
            assert (
                len(set(sizes)) == 1
            ), f"Data Asset Onboarding Error: All data must have the same number of rows. Lengths: {str(set(sizes))}"

            dataset_payload = onboard_functions.build_payload_dataset(
                dataset, input_ref_data, output_ref_data, hypotheses_ref_data, meta_ref_data, tags, error_handler
            )
            assert (
                len(dataset_payload["input"]) > 0
            ), "Data Asset Onboarding Error: Please specify the input data of your dataset."
            # assert (
            #     len(dataset_payload["output"]) > 0
            # ), "Data Asset Onboarding Error: Please specify the output data of your dataset."

            response = onboard_functions.create_data_asset(payload=dataset_payload, data_asset_type="dataset", api_key=api_key)
            if response["success"] is True:
                return_dict = {"status": response["status"], "asset_id": response["asset_id"]}
            else:
                raise Exception(response["error"])
            shutil.rmtree(folder)
            if csv_path is not None and os.path.exists(csv_path) is True:
                os.remove(csv_path)
        except Exception as e:
            if folder is not None:
                shutil.rmtree(folder)
            if csv_path is not None and os.path.exists(csv_path) is True:
                os.remove(csv_path)
            raise Exception(e)
        return return_dict



__author__ = "michaellam"

from pathlib import Path
import json
from aixplain.factories.model_factory import ModelFactory
from tests.test_utils import delete_asset, delete_service_account
from aixplain.utils import config
import docker
import pytest


def test_create_and_upload_model():
    # List the host machines
    host_response = ModelFactory.list_host_machines()
    for hosting_machine_dict in host_response:
        assert "code" in hosting_machine_dict.keys()
        assert "type" in hosting_machine_dict.keys()
        assert "cores" in hosting_machine_dict.keys()
        assert "memory" in hosting_machine_dict.keys()
        assert "hourlyCost" in hosting_machine_dict.keys()

    # List the functions
    response = ModelFactory.list_functions()
    items = response["items"]
    for item in items:
        assert "output" not in item.keys()
        assert "params" not in item.keys()
        assert "id" not in item.keys()
        assert "name" in item.keys()

    # Register the model, and create an image repository for it.
    with open(Path("tests/test_requests/create_asset_request.json")) as f:
        mock_register_payload = json.load(f)
    name = mock_register_payload["name"]
    description = mock_register_payload["description"]
    function = mock_register_payload["function"]
    source_language = mock_register_payload["sourceLanguage"]
    input_modality = mock_register_payload["input_modality"]
    output_modality = mock_register_payload["output_modality"]
    documentation_url = mock_register_payload["documentation_url"]
    register_response = ModelFactory.create_asset_repo(name, description, function, source_language, input_modality, output_modality, documentation_url, config.TEAM_API_KEY)
    assert "id" in register_response.keys()
    assert "repositoryName" in register_response.keys()
    model_id = register_response["id"]
    repo_name = register_response["repositoryName"]

    # Log into the image repository.
    login_response = ModelFactory.asset_repo_login()

    assert login_response["username"] == "AWS"
    assert login_response["registry"] == "535945872701.dkr.ecr.us-east-1.amazonaws.com"
    assert "password" in login_response.keys()

    username = login_response["username"]
    password = login_response["password"]
    registry = login_response["registry"]

    # Push an image to ECR
    low_level_client = docker.APIClient(base_url="unix://var/run/docker.sock")
    low_level_client.pull("bash")
    low_level_client.tag("bash", f"{registry}/{repo_name}")
    low_level_client.push(f"{registry}/{repo_name}", auth_config={"username": username, "password": password})

    # Send an email to finalize onboarding process
    ModelFactory.onboard_model(model_id, "latest", "fake_hash")

    # Clean up
    delete_service_account(config.TEAM_API_KEY)
    delete_asset(model_id, config.TEAM_API_KEY)



__author__ = "thiagocastroferreira"

import aixplain.utils.config as config
from aixplain.enums import Function, Supplier
from aixplain.enums.asset_status import AssetStatus
from aixplain.modules.agent import Agent, ModelTool, PipelineTool
from typing import Dict, Text
from urllib.parse import urljoin


def build_agent(payload: Dict, api_key: Text = config.TEAM_API_KEY) -> Agent:
    """Instantiate a new agent in the platform."""
    tools = payload["assets"]
    for i, tool in enumerate(tools):
        if tool["type"] == "model":
            for supplier in Supplier:
                if tool["supplier"] is not None and tool["supplier"].lower() in [
                    supplier.value["code"].lower(),
                    supplier.value["name"].lower(),
                ]:
                    tool["supplier"] = supplier
                    break

            tool = ModelTool(
                function=Function(tool["function"]) if tool["function"] is not None else None,
                supplier=tool["supplier"],
                version=tool["version"],
                model=tool["assetId"],
            )
        elif tool["type"] == "pipeline":
            tool = PipelineTool(description=tool["description"], pipeline=tool["assetId"])
        else:
            raise Exception("Agent Creation Error: Tool type not supported.")
        tools[i] = tool

    agent = Agent(
        id=payload["id"],
        name=payload["name"] if "name" in payload else "",
        tools=tools,
        description=payload["description"] if "description" in payload else "",
        supplier=payload["teamId"] if "teamId" in payload else None,
        version=payload["version"] if "version" in payload else None,
        cost=payload["cost"] if "cost" in payload else None,
        llm_id=payload["llmId"] if "llmId" in payload else "6646261c6eb563165658bbb1",
        api_key=api_key,
        status=AssetStatus(payload["status"]),
    )
    agent.url = urljoin(config.BACKEND_URL, f"sdk/agents/{agent.id}/run")
    return agent


def validate_llm(model_id: Text) -> None:
    from aixplain.factories.model_factory import ModelFactory

    try:
        llm = ModelFactory.get(model_id)
        assert llm.function == Function.TEXT_GENERATION, "Large Language Model must be a text generation model."
    except Exception:
        raise Exception(f"Large Language Model with ID '{model_id}' not found.")



__author__ = "lucaspavanelli"
"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import uuid
import time
import json
from dotenv import load_dotenv

load_dotenv()
from aixplain.factories import ModelFactory
from aixplain.factories import DatasetFactory
from aixplain.factories import FinetuneFactory
from aixplain.modules.finetune.cost import FinetuneCost
from aixplain.enums import Function, Language
from datetime import datetime, timedelta, timezone

import pytest

TIMEOUT = 20000.0
RUN_FILE = "tests/functional/finetune/data/finetune_test_end2end.json"
ESTIMATE_COST_FILE = "tests/functional/finetune/data/finetune_test_cost_estimation.json"
LIST_FILE = "tests/functional/finetune/data/finetune_test_list_data.json"
PROMPT_FILE = "tests/functional/finetune/data/finetune_test_prompt_validator.json"


def read_data(data_path):
    return json.load(open(data_path, "r"))


@pytest.fixture(scope="module", params=read_data(ESTIMATE_COST_FILE))
def estimate_cost_input_map(request):
    return request.param


@pytest.fixture(scope="module", params=read_data(LIST_FILE))
def list_input_map(request):
    return request.param


@pytest.fixture(scope="module", params=read_data(PROMPT_FILE))
def validate_prompt_input_map(request):
    return request.param


def pytest_generate_tests(metafunc):
    if "input_map" in metafunc.fixturenames:
        four_weeks_ago = datetime.now(timezone.utc) - timedelta(weeks=4)
        models = ModelFactory.list(function=Function.TEXT_GENERATION, is_finetunable=True)["results"]

        recent_models = [
            {
                "model_name": model.name,
                "model_id": model.id,
                "dataset_name": "Test text generation dataset",
                "inference_data": "Hello!",
                "required_dev": True,
                "search_metadata": False,
            }
            for model in models
            if model.created_at is not None and model.created_at >= four_weeks_ago
        ]
        recent_models += read_data(RUN_FILE)
        metafunc.parametrize("input_map", recent_models)


def test_end2end(input_map):
    model = input_map["model_id"]
    dataset_list = [DatasetFactory.list(query=input_map["dataset_name"])["results"][0]]
    train_percentage, dev_percentage = 100, 0
    if input_map["required_dev"]:
        train_percentage, dev_percentage = 80, 20
    finetune = FinetuneFactory.create(
        str(uuid.uuid4()), dataset_list, model, train_percentage=train_percentage, dev_percentage=dev_percentage
    )
    assert type(finetune.cost) is FinetuneCost
    cost_map = finetune.cost.to_dict()
    assert "trainingCost" in cost_map
    assert "hostingCost" in cost_map
    assert "inferenceCost" in cost_map
    finetune_model = finetune.start()
    start, end = time.time(), time.time()
    status = finetune_model.check_finetune_status().model_status.value
    while status != "onboarded" and (end - start) < TIMEOUT:
        status = finetune_model.check_finetune_status().model_status.value
        assert status != "failed"
        time.sleep(5)
        end = time.time()
    assert finetune_model.check_finetune_status().model_status.value == "onboarded"
    time.sleep(30)
    print(f"Model dict: {finetune_model.__dict__}")
    result = finetune_model.run(input_map["inference_data"])
    print(f"Result: {result}")
    assert result is not None
    if input_map["search_metadata"]:
        assert "details" in result
        assert len(result["details"]) > 0
        assert "metadata" in result["details"][0]
        assert len(result["details"][0]["metadata"]) > 0
    finetune_model.delete()


def test_cost_estimation_text_generation(estimate_cost_input_map):
    model = ModelFactory.get(estimate_cost_input_map["model_id"])
    dataset_list = [DatasetFactory.list(query=estimate_cost_input_map["dataset_name"])["results"][0]]
    finetune = FinetuneFactory.create(str(uuid.uuid4()), dataset_list, model)
    assert type(finetune.cost) is FinetuneCost
    cost_map = finetune.cost.to_dict()
    assert "trainingCost" in cost_map
    assert "hostingCost" in cost_map
    assert "inferenceCost" in cost_map


def test_list_finetunable_models(list_input_map):
    model_list = ModelFactory.list(
        function=Function(list_input_map["function"]),
        source_languages=Language(list_input_map["source_language"]) if "source_language" in list_input_map else None,
        target_languages=Language(list_input_map["target_language"]) if "target_language" in list_input_map else None,
        is_finetunable=True,
    )["results"]
    assert len(model_list) > 0


def test_prompt_validator(validate_prompt_input_map):
    model = ModelFactory.get(validate_prompt_input_map["model_id"])
    dataset_list = [DatasetFactory.list(query=validate_prompt_input_map["dataset_name"])["results"][0]]
    if validate_prompt_input_map["is_valid"]:
        finetune = FinetuneFactory.create(
            str(uuid.uuid4()), dataset_list, model, prompt_template=validate_prompt_input_map["prompt_template"]
        )
        assert finetune is not None
    else:
        with pytest.raises(Exception) as exc_info:
            finetune = FinetuneFactory.create(
                str(uuid.uuid4()), dataset_list, model, prompt_template=validate_prompt_input_map["prompt_template"]
            )
        assert exc_info.type is AssertionError



__author__ = "shreyassharma"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: May 15th 2023
Description:
    Data Factory Class
"""

import aixplain.utils.config as config
import logging

from aixplain.factories.asset_factory import AssetFactory
from aixplain.modules.data import Data
from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.function import Function
from aixplain.enums.language import Language
from aixplain.enums.license import License
from aixplain.enums.privacy import Privacy
from aixplain.utils.file_utils import _request_with_retry
from aixplain.utils import config
from typing import Any, Dict, List, Text
from urllib.parse import urljoin
from uuid import uuid4


class DataFactory(AssetFactory):
    """A static class for creating and exploring Dataset Objects.

    Attributes:
        backend_url (str): The URL for the backend.
    """

    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def __from_response(cls, response: Dict) -> Data:
        """Converts Json response to 'Data' object

        Args:
            response (dict): Json from API

        Returns:
            Data: Converted 'Data' object
        """
        languages = []
        if "languages" in response["metadata"]:
            languages = []
            for lng in response["metadata"]["languages"]:
                if "dialect" not in lng:
                    lng["dialect"] = ""
                languages.append(Language(lng))

        data = Data(
            id=response["id"],
            name=response["name"],
            dtype=DataType(response["dataType"]),
            dsubtype=DataSubtype(response["dataSubtype"]),
            privacy=Privacy.PRIVATE,
            languages=languages,
            onboard_status=response["status"],
            length=int(response["segmentsCount"])
            if "segmentsCount" in response and response["segmentsCount"] is not None
            else None,
        )
        return data

    @classmethod
    def get(cls, data_id: Text) -> Data:
        """Create a 'Data' object from dataset id

        Args:
            data_id (Text): Data ID of required dataset.

        Returns:
            Data: Created 'Data' object
        """
        url = urljoin(cls.backend_url, f"sdk/data/{data_id}/overview")
        if cls.aixplain_key != "":
            headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        logging.info(f"Start service for GET Data  - {url} - {headers}")
        r = _request_with_retry("get", url, headers=headers)
        resp = r.json()
        if "statusCode" in resp and resp["statusCode"] == 404:
            raise Exception(f"Data GET Error: Data {data_id} not found.")
        return cls.__from_response(resp)



__author__ = "shreyassharma"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: October 28th 2022
Description:
    Datasets Class
"""

import logging

from aixplain.enums.function import Function
from aixplain.enums.license import License
from aixplain.enums.onboard_status import OnboardStatus
from aixplain.enums.privacy import Privacy
from aixplain.modules.asset import Asset
from aixplain.modules.data import Data
from aixplain.utils import config
from aixplain.utils.file_utils import _request_with_retry
from urllib.parse import urljoin
from typing import Any, Dict, List, Optional, Text


class Dataset(Asset):
    """Dataset is a collection of data intended to be used for a specific function.
            Different from corpus, a dataset is a representative sample of a specific phenomenon to a specific AI task.
            aiXplain also counts with an extensive collection of datasets for training, infer and benchmark various tasks like
            Translation, Speech Recognition, Diacritization, Sentiment Analysis, and much more.

    Attributes:
        id (Text): Dataset ID
        name (Text): Dataset Name
        description (Text): Dataset description
        function (Function): Function for which the dataset is intented to
        source_data (Dict[Any, Data]): List of input Data to the function
        target_data (Dict[Any, List[Data]]): List of Multi-reference Data which is expected to be outputted by the function
        onboard_status (OnboardStatus): onboard status
        hypotheses (Dict[Any, Data], optional): dataset's hypotheses, i.e. model outputs based on the source data. Defaults to {}.
        metadata (Dict[Any, Data], optional): dataset's metadata. Defaults to {}.
        tags (List[Text], optional): tags that describe the dataset. Defaults to [].
        license (Optional[License], optional): Dataset License. Defaults to None.
        privacy (Privacy, optional): Dataset Privacy. Defaults to Privacy.PRIVATE.
        supplier (Text, optional): Dataset Supplier. Defaults to "aiXplain".
        version (Text, optional): Dataset Version. Defaults to "1.0".
    """

    def __init__(
        self,
        id: Text,
        name: Text,
        description: Text,
        function: Function,
        source_data: Dict[Any, Data],
        target_data: Dict[Any, List[Data]],
        onboard_status: OnboardStatus,
        hypotheses: Dict[Any, Data] = {},
        metadata: Dict[Any, Data] = {},
        tags: List[Text] = [],
        license: Optional[License] = None,
        privacy: Privacy = Privacy.PRIVATE,
        supplier: Text = "aiXplain",
        version: Text = "1.0",
        length: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Dataset Class.

        Description:
            Dataset is a collection of data intended to be used for a specific function.
            Different from corpus, a dataset is a representative sample of a specific phenomenon to a specific AI task.
            aiXplain also counts with an extensive collection of datasets for training, infer and benchmark various tasks like
            Translation, Speech Recognition, Diacritization, Sentiment Analysis, and much more.

        Args:
            id (Text): Dataset ID
            name (Text): Dataset Name
            description (Text): Dataset description
            function (Function): Function for which the dataset is intented to
            source_data (Dict[Any, Data]): List of input Data to the function
            target_data (Dict[Any, List[Data]]): List of Multi-reference Data which is expected to be outputted by the function
            onboard_status (OnboardStatus): onboard status
            hypotheses (Dict[Any, Data], optional): dataset's hypotheses, i.e. model outputs based on the source data. Defaults to {}.
            metadata (Dict[Any, Data], optional): dataset's metadata. Defaults to {}.
            tags (List[Text], optional): tags that describe the dataset. Defaults to [].
            license (Optional[License], optional): Dataset License. Defaults to None.
            privacy (Privacy, optional): Dataset Privacy. Defaults to Privacy.PRIVATE.
            supplier (Text, optional): Dataset Supplier. Defaults to "aiXplain".
            version (Text, optional): Dataset Version. Defaults to "1.0".
            length (Optional[int], optional): Number of rows in the Dataset. Defaults to None.
        """
        super().__init__(
            id=id, name=name, description=description, supplier=supplier, version=version, license=license, privacy=privacy
        )
        if isinstance(onboard_status, str):
            onboard_status = OnboardStatus(onboard_status)
        self.onboard_status = onboard_status
        self.function = function
        self.source_data = source_data
        self.target_data = target_data
        self.hypotheses = hypotheses
        self.metadata = metadata
        self.tags = tags
        self.length = length
        self.kwargs = kwargs

    def __repr__(self):
        return f"<Dataset: {self.name}>"

    def delete(self) -> None:
        """Delete Dataset service"""
        try:
            url = urljoin(config.BACKEND_URL, f"sdk/datasets/{self.id}")
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
            logging.info(f"Start service for DELETE Dataset  - {url} - {headers}")
            r = _request_with_retry("delete", url, headers=headers)
            if r.status_code != 200:
                raise Exception()
        except Exception:
            message = "Dataset Deletion Error: Make sure the dataset exists and you are the owner."
            logging.error(message)
            raise Exception(f"{message}")



# Copyright 2023 aiXplain authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Python implementation of the aiXplain speech transcription streaming client."""

import argparse
import sys
import time
from typing import List
import grpc
import logging

import aixplain_speech_transcription_streaming_pb2 as pb
import aixplain_speech_transcription_streaming_pb2_grpc

FRAME_RATE = 16000
FFMPEG_FORMAT = 'wav'

PS_SUBTITLE = 'subtitle'
PS_LOG = 'log'
RED = "\033[0;31m"
GREEN = "\033[0;32m"
DEFAULT = "\033[0;97m"

def grpc_duration_to_seconds(duration):
    seconds = float(duration.seconds)
    nanos = float(duration.nanos) / 1e9
    return seconds + nanos

def generate_payloads(file_path):
    # uncomment this if your audio file is not compatible
    stream_configuration = pb.SpeechRecognitionRequest(
        config=pb.AudioConfig(encoding="LINEAR16", hertz=FRAME_RATE, language_code="en"),
    )
    yield stream_configuration
    # Iterate over the raw bytes in chunks
    chunk_size = 16000 # half a second of audio
    chunk_period = 0.5
    logging.info(f'Sending chunks...')
    with open(file_path, "rb") as audio_file:
        i = 0
        while True:
            chunk = audio_file.read(chunk_size)
            if not chunk:
                break
            time.sleep((len(chunk)/chunk_size)*chunk_period) # simulate streaming by introducing sleep
            logging.debug(f'Sending chunk {i}')
            payload = pb.SpeechRecognitionRequest(audio_content=chunk)
            yield payload
            i += 1

def consume_results(response: List[pb.SpeechRecognitionResponse], print_style):
    for i, inference in enumerate(response):
        if i == 0:
            logging.info(f'Detected language {inference.language_code=}')
        if len(inference.hypotheses):
            # get the top hypothesis
            hypothesis = inference.hypotheses[0]
            transcript, confidence = hypothesis.transcript, hypothesis.confidence
            if inference.is_final:
                logging.info(f't={grpc_duration_to_seconds(inference.end_time):.3f}  conf={confidence:.2f} FINAL="{transcript}"')
                if print_style == PS_SUBTITLE:
                    sys.stdout.write(GREEN)
                    sys.stdout.write("\033[K")
                    sys.stdout.write(f'{grpc_duration_to_seconds(inference.end_time):.3f}' + ": " + transcript + "\n")
            else:
                logging.info(f't={grpc_duration_to_seconds(inference.end_time):.3f}  conf={confidence:.2f} chunk="{transcript}"')
                if print_style == PS_SUBTITLE:
                    sys.stdout.write(RED)
                    sys.stdout.write("\033[K")
                    sys.stdout.write(f'{grpc_duration_to_seconds(inference.end_time):.3f}' + ": " + transcript + "\r")

            for word in hypothesis.words:
                logging.info(f'Word: {word.word.ljust(12)} '
                    f'Start-End: {grpc_duration_to_seconds(word.start_time):.3f}-{grpc_duration_to_seconds(word.end_time):.3f} '
                    f'Confidence: {word.confidence:.3f} ')
        else:  # called at the end
            logging.info(f'{inference.is_final=} server processing_time={grpc_duration_to_seconds(inference.end_time):.3f}')
            if print_style == PS_SUBTITLE:
                sys.stdout.write(DEFAULT)
                sys.stdout.write("Exiting...\n")

def _stream_file(channel, file_path, print_style):
    stub = aixplain_speech_transcription_streaming_pb2_grpc.AixplainSpeechStreamingStub(channel)
    response = stub.SpeechRecognize(generate_payloads(file_path))
    consume_results(response, print_style)

def run_insecure(host, file_path, print_style):
    with grpc.insecure_channel(host, options=(('grpc.ssl_target_name_override', host),)) as channel:
        _stream_file(channel, file_path, print_style)

def run(host, client_cert, client_key, ca_cert, file_path, print_style):
    def create_secure_channel(host, client_cert, client_key, ca_cert):
        with open(client_cert, 'rb') as f:
            client_cert_data = f.read()

        with open(client_key, 'rb') as f:
            client_key_data = f.read()

        with open(ca_cert, 'rb') as f:
            ca_cert_data = f.read()

        credentials = grpc.ssl_channel_credentials(
            root_certificates=ca_cert_data,
            private_key=client_key_data,
            certificate_chain=client_cert_data
        )
        return grpc.secure_channel(host, credentials)
    with create_secure_channel(host, client_cert, client_key, ca_cert) as channel:
        _stream_file(channel, file_path, print_style)


if __name__ == "__main__":
    log_format = "%(asctime)s - %(levelname)-7s - P%(process)d: %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)
    parser = argparse.ArgumentParser(description="aiXplain speech recognition streaming client.")

    parser.add_argument('--print-style', default='log', choices=[PS_SUBTITLE, PS_LOG], help='The print style, either "log" or "subtitle"')
    parser.add_argument('--addr', default='localhost:50051', help='the address to connect to (default "localhost:50051")')
    parser.add_argument('--cacert', default='./client-crt/ca.crt', help='ca cert for mTLS (default "./client-crt/ca.crt")')
    parser.add_argument('--cert', default='./client-crt/tls.crt', help='client cert for mTLS (default "./client-crt/tls.crt")')
    parser.add_argument('--key', default='./client-crt/tls.key', help='client key for mTLS (default "./client-crt/tls.key")')
    parser.add_argument('--insecure', action='store_true', help='use insecure connection (no mTLS)')
    parser.add_argument('--file-path', default="resources/conv.wav", help='audio file to stream from')

    args = parser.parse_args()

    if args.print_style == PS_SUBTITLE:
        logging.getLogger('').setLevel(logging.ERROR)

    if args.insecure:
        run_insecure(args.addr, args.file_path, args.print_style)
    else:
        run(args.addr, args.cert, args.key, args.cacert, args.file_path,  args.print_style)



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: December 27th 2022
Description:
    Asset Factory Class
"""

from abc import abstractmethod
from typing import List, Text
from aixplain.modules.asset import Asset
from aixplain.utils import config


class AssetFactory:
    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @abstractmethod
    def get(self, asset_id: Text) -> Asset:
        """Create a 'Asset' object from id

        Args:
            asset_id (str): ID of required asset.

        Returns:
            Asset: Created 'Asset' object
        """
        pass



__author__ = "aiXplain"

"""
Copyright 2023 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: aiXplain team
Date: November 22nd 2023
Description:
    Asset Ownership Type
"""

from enum import Enum


class OwnershipType(Enum):
    SUBSCRIBED = "SUBSCRIBED"
    OWNED = "OWNED"

    def __str__(self):
        return self._value_



"""
aiXplain SDK Library.
---

aiXplain SDK enables python programmers to add AI functions
to their software.

Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Set default logging handler to avoid "No handler found" warnings.
import os
import logging
from logging import NullHandler

LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO").upper()
logging.basicConfig(level=LOG_LEVEL)
logging.getLogger(__name__).addHandler(NullHandler())

# Load Environment Variables
from dotenv import load_dotenv

load_dotenv()

# Validate API keys
from aixplain.utils import config

if config.TEAM_API_KEY == "" and config.AIXPLAIN_API_KEY == "":
    raise Exception(
        "'TEAM_API_KEY' has not been set properly and is empty. For help, please refer to the documentation (https://github.com/aixplain/aixplain#api-key-setup)"
    )



from .nodes import (
    AssetNode,
    Decision,
    Script,
    Input,
    Output,
    Route,
    Router,
    BaseReconstructor,
    BaseSegmentor,
)
from .pipeline import DesignerPipeline
from .base import (
    Node,
    Link,
    Param,
    ParamProxy,
    InputParam,
    OutputParam,
    Inputs,
    Outputs,
    TI,
    TO,
)
from .enums import (
    ParamType,
    RouteType,
    Operation,
    NodeType,
    AssetType,
    FunctionType,
)
from .mixins import LinkableMixin, OutputableMixin, RoutableMixin


__all__ = [
    "DesignerPipeline",
    "AssetNode",
    "Decision",
    "Script",
    "Input",
    "Output",
    "Route",
    "Router",
    "BaseReconstructor",
    "BaseSegmentor",
    "Node",
    "Link",
    "Param",
    "ParamType",
    "InputParam",
    "OutputParam",
    "RouteType",
    "Operation",
    "NodeType",
    "AssetType",
    "FunctionType",
    "LinkableMixin",
    "OutputableMixin",
    "RoutableMixin",
    "Inputs",
    "Outputs",
    "ParamProxy",
    "TI",
    "TO",
]



__author__ = "aiXplain"

import aixplain.processes.data_onboarding.process_media_files as process_media_files
import aixplain.processes.data_onboarding.process_text_files as process_text_files
import aixplain.utils.config as config
import logging
import os
import pandas as pd
import random

from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.error_handler import ErrorHandler
from aixplain.enums.file_type import FileType
from aixplain.enums.storage_type import StorageType
from aixplain.modules.corpus import Corpus
from aixplain.modules.data import Data
from aixplain.modules.dataset import Dataset
from aixplain.modules.file import File
from aixplain.modules.metadata import MetaData
from aixplain.utils.file_utils import _request_with_retry
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Text, Union
from urllib.parse import urljoin

FORBIDDEN_COLUMN_NAMES = [
    "@VOLUME",
    "@START_TIME",
    "@END_TIME",
    "@START",
    "@END",
    "@ORIGINAL",
    "@SOURCE",
    "@INDEX",
    "@SPLIT",
    "@STATUS",
]


def get_paths(input_paths: List[Union[str, Path]]) -> List[Path]:
    """Recursively access all local paths. Check if file extensions are supported.

    Args:
        input_paths (List[Union[str, Path]]): list of input pahts including folders and files

    Returns:
        List[Path]: list of local file paths
    """
    paths = []
    for path in input_paths:
        if isinstance(path, str):
            path = Path(path)

        if path.is_dir():
            for subpath in path.iterdir():
                extension = subpath.suffix
                if extension == FileType.CSV:
                    paths.append(subpath)
                else:
                    logging.warning(f"Data Asset Onboarding: File Extension not Supported ({str(path)})")
        else:
            extension = FileType(path.suffix)
            if extension == FileType.CSV:
                paths.append(path)
            else:
                logging.warning(f"Data Asset Onboarding: File Extension not Supported ({str(path)})")

    # check CSV sizes
    for path in paths:
        assert os.path.getsize(path) <= 1e9, f'Data Asset Onboarding Error: CSV file "{path}" exceeds the size limit of 1 GB.'
    return paths


def process_data_files(
    data_asset_name: str, metadata: MetaData, paths: List, folder: Optional[Union[str, Path]] = None
) -> Tuple[List[File], int, int, int, int]:
    """Process a list of local files, compress and upload them to pre-signed URLs in S3

    Args:
        data_asset_name (str): name of the data asset
        metadata (MetaData): meta data of the asset
        paths (List): list of paths to local files
        folder (Union[str, Path], optional): local folder to save compressed files before upload them to s3. Defaults to data_asset_name.

    Returns:
        Tuple[List[File], int, int, int]: list of s3 links; data, start and end columns index; and number of rows
    """
    if folder is None:
        folder = Path(data_asset_name)
    elif isinstance(folder, str):
        folder = Path(folder)

    files = []
    data_column_idx, start_column_idx, end_column_idx, nrows, = (
        -1,
        -1,
        -1,
        0,
    )
    if metadata.dtype in [DataType.AUDIO, DataType.IMAGE, DataType.LABEL] or metadata.dsubtype == DataSubtype.INTERVAL:
        files, data_column_idx, start_column_idx, end_column_idx, nrows = process_media_files.run(
            metadata=metadata, paths=paths, folder=folder
        )
    elif metadata.dtype in [DataType.TEXT]:
        files, data_column_idx, nrows = process_text_files.run(metadata=metadata, paths=paths, folder=folder)
    return files, data_column_idx, start_column_idx, end_column_idx, nrows


def build_payload_data(data: Data) -> Dict:
    """Create data payload to call coreengine on Corpus/Dataset onboard

    Args:
        data (Data): data object

    Returns:
        Dict: payload
    """
    data_json = {
        "name": data.name,
        "dataColumn": data.data_column,
        "dataType": data.dtype.value,
        "dataSubtype": data.dsubtype.value,
        "batches": [{"tempFilePath": str(file.path), "order": idx + 1} for idx, file in enumerate(data.files)],
        "tags": [],
        "metaData": {"languages": []},
    }

    if len(data.languages) > 0:
        data_json["metaData"]["languages"] = [lng.value for lng in data.languages]

    if data.start_column is not None and data.start_column > -1 and data.end_column is not None and data.end_column > -1:
        if data.dtype == DataType.AUDIO:
            data_json["startTimeColumn"] = data.start_column
            data_json["endTimeColumn"] = data.end_column
        else:
            data_json["startColumn"] = data.start_column
            data_json["endColumn"] = data.end_column
    return data_json


def build_payload_corpus(corpus: Corpus, ref_data: List[Text], error_handler: ErrorHandler) -> Dict:
    """Create corpus payload to call coreengine on the onboard process

    Args:
        corpus (Corpus): corpus object
        ref_data (List[Text]): list of referred data
        error_handler (ErrorHandler): how to handle failed rows

    Returns:
        Dict: payload
    """
    payload = {
        "name": corpus.name,
        "description": corpus.description,
        "suggestedFunctions": [f.value for f in corpus.functions],
        "onboardingErrorsPolicy": error_handler.value,
        "tags": corpus.tags,
        "pricing": {"usage": ["benchmark", "finetune"], "type": "free", "cost": 0},
        "privacy": corpus.privacy.value,
        "license": {"typeId": corpus.license.value},
        "refData": ref_data,
        "data": [],
    }

    for data in corpus.data:
        data_json = build_payload_data(data)
        payload["data"].append(data_json)
    return payload


def build_payload_dataset(
    dataset: Dataset,
    input_ref_data: Dict[Text, Any],
    output_ref_data: Dict[Text, List[Any]],
    hypotheses_ref_data: Dict[Text, Any],
    meta_ref_data: Dict[Text, Any],
    tags: List[Text],
    error_handler: ErrorHandler,
) -> Dict:
    """Generate onboard payload to coreengine

    Args:
        dataset (Dataset): dataset to be onboard
        input_ref_data (Dict[Text, Any]): reference to existent input data
        output_ref_data (Dict[Text, List[Any]]): reference to existent output data
        hypotheses_ref_data (Dict[Text, Any]): reference to existent hypotheses to the target data
        meta_ref_data (Dict[Text, Any]): reference to existent metadata
        tags (List[Text]): description tags
        error_handler (ErrorHandler): how to handle failed rows

    Returns:
        Dict: onboard payload
    """
    # compute ref data
    flat_input_ref_data = list(input_ref_data.values())
    flat_output_ref_data = [item for sublist in list(output_ref_data.values()) for item in sublist]
    flat_meta_ref_data = list(meta_ref_data.values())
    flat_hypotheses_ref_data = list(hypotheses_ref_data.values())
    ref_data = flat_input_ref_data + flat_output_ref_data + flat_meta_ref_data + flat_hypotheses_ref_data

    payload = {
        "name": dataset.name,
        "description": dataset.description,
        "function": dataset.function.value,
        "onboardingErrorsPolicy": error_handler.value,
        "tags": dataset.tags,
        "privacy": dataset.privacy.value,
        "license": {"typeId": dataset.license.value},
        "refData": ref_data,
        "tags": tags,
        "data": [],
        "input": [],
        "hypotheses": [],
        "metadata": [],
    }

    # INPUT DATA
    # compute ref input data
    index = 1
    for data_name in dataset.source_data:
        data = dataset.source_data[data_name]
        data_json = build_payload_data(data)
        data_json["tempId"] = data.id
        payload["data"].append(data_json)
        payload["input"].append({"index": index, "name": data.name, "dataId": data.id})
        index += 1

    # process input ref data
    for data_name in input_ref_data:
        payload["input"].append({"index": index, "name": data_name, "dataId": input_ref_data[data_name]})
        index += 1

    # HYPOTHESES
    index = 1
    for data_name in dataset.hypotheses:
        data = dataset.hypotheses[data_name]
        data_json = build_payload_data(data)
        data_json["tempId"] = data.id
        payload["data"].append(data_json)
        payload["hypotheses"].append({"index": index, "name": data.name, "dataId": data.id})
        index += 1

    # process hypotheses ref data
    for data_name in hypotheses_ref_data:
        payload["hypotheses"].append({"index": index, "name": data_name, "dataId": hypotheses_ref_data[data_name]})
        index += 1

    # METADATA
    index = 1
    for data_name in dataset.metadata:
        data = dataset.metadata[data_name]
        data_json = build_payload_data(data)
        data_json["tempId"] = data.id
        payload["data"].append(data_json)
        payload["metadata"].append({"index": index, "name": data.name, "dataId": data.id})
        index += 1

    # process meta ref data
    for data_name in meta_ref_data:
        payload["metadata"].append({"index": index, "name": data_name, "dataId": meta_ref_data[data_name]})
        index += 1

    # OUTPUT DATA
    # compute output data
    output = {}
    index = 1
    for output_name in dataset.target_data:
        output_data = {"index": index, "name": output_name, "dataIds": []}
        for data in dataset.target_data[output_name]:
            data_json = build_payload_data(data)
            data_json["tempId"] = data.id
            payload["data"].append(data_json)
            output_data["dataIds"].append(data.id)
        output[output_name] = output_data
        index += 1

    # process output ref data
    for output_name in output_ref_data:
        if output_name in output:
            output[output_name]["dataIds"] += output_ref_data[output_name]
        else:
            output_data = {"index": index, "name": output_name, "dataIds": []}
            for data_id in output_ref_data[output_name]:
                output_data["dataIds"].append(data_id)
            index += 1
            output[output_name] = output_data
    payload["output"] = list(output.values())
    return payload


def create_data_asset(payload: Dict, data_asset_type: Text = "corpus", api_key: Optional[Text] = None) -> Dict:
    """Service to call onboard process in coreengine

    Args:
        payload (Dict): onboard payload
        data_asset_type (Text, optional): corpus or dataset. Defaults to "corpus".
        api_key (Optional[Text]): team api key. Defaults to None.

    Returns:
        Dict: onboard status
    """
    if api_key is not None:
        team_key = api_key
    else:
        team_key = config.TEAM_API_KEY
    headers = {"Authorization": "token " + team_key}

    url = urljoin(config.BACKEND_URL, f"sdk/{data_asset_type}/onboard")

    r = _request_with_retry("post", url, headers=headers, json=payload)
    if 200 <= r.status_code < 300:
        response = r.json()

        asset_id = response["id"]
        status = response["status"]

        return {
            "success": True,
            "asset_id": asset_id,
            "status": status,
            # "coreengine_payload": coreengine_payload
        }
    else:
        try:
            response = r.json()
            msg = response["message"]
            error_msg = f"Data Asset Onboarding Error: {msg}"
        except Exception as e:
            error_msg = (
                f"Data Asset Onboarding Error: Failure on creating the {data_asset_type}. Please contant the administrators."
            )
        return {"success": False, "error": error_msg}


def is_data(data_id: Text) -> bool:
    """Check whether reference data exists

    Args:
        data_id (Text): ID of the data

    Returns:
        bool: True if it exists, False otherwise
    """
    try:
        api_key = config.TEAM_API_KEY
        backend_url = config.BACKEND_URL
        url = urljoin(backend_url, f"sdk/data/{data_id}/overview")
        headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
        r = _request_with_retry("get", url, headers=headers)
        resp = r.json()

        if "id" in resp:
            return True
        return False
    except:
        return False


def split_data(paths: List, split_rate: List[float], split_labels: List[Text]) -> MetaData:
    """Split the data according to some split labels and rate

    Args:
        paths (List): path to data files
        split_rate (List[Text]): split rate
        split_labels (List[Text]): split labels

    Returns:
        MetaData: metadata of the new split
    """
    # get column name
    column_name = None
    for path in paths:
        try:
            dataframe = pd.read_csv(path)
            for candidate_name in ["split", "SPLIT", "_split", "_split_", "split_"]:
                if candidate_name not in dataframe.columns:
                    column_name = candidate_name
                    break

            if column_name is not None:
                break
        except Exception as e:
            message = f'Data Asset Onboarding Error: Local file "{path}" not found.'
            logging.exception(message)
            raise Exception(message)

    if column_name is None:
        message = f"Data Asset Onboarding Error: All split names are used."
        raise Exception(message)

    for path in paths:
        dataframe = pd.read_csv(path)
        dataframe[column_name] = [slabel for (slabel, srate) in zip(split_labels, split_rate) if srate == max(split_rate)][0]

        size = len(dataframe)
        start = 0
        indexes = list(range(size))
        random.shuffle(indexes)
        for (slabel, srate) in zip(split_labels, split_rate):
            split_size = int(srate * size)

            split_indexes = indexes[start : start + split_size]
            dataframe.loc[split_indexes, column_name] = slabel
            start = start + split_size
        dataframe.to_csv(path, index=False)

    return MetaData(name=column_name, dtype=DataType.LABEL, dsubtype=DataSubtype.SPLIT, storage_type=StorageType.TEXT)



"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import requests

import aixplain.utils.config as config
from aixplain.enums.license import License

from collections import defaultdict
from pathlib import Path
from requests.adapters import HTTPAdapter, Retry
from typing import Any, Optional, Text, Union, Dict, List
from uuid import uuid4
from urllib.parse import urljoin, urlparse
from pandas import DataFrame


def save_file(download_url: Text, download_file_path: Optional[Any] = None) -> Any:
    """Download and save file from given URL

    Args:
        download_url (Text): URL of file to download
        download_file_path (Any, optional): File path to save downloaded file. If None then generates a folder 'aiXplain' in current working directory. Defaults to None.

    Returns:
        Text: Path where file was downloaded
    """
    if download_file_path is None:
        save_dir = os.getcwd()
        save_dir = Path(save_dir) / "aiXplain"
        save_dir.mkdir(parents=True, exist_ok=True)
        file_ext = Path(download_url).suffix.split("?")[0]
        download_file_path = save_dir / (str(uuid4()) + file_ext)
    r = _request_with_retry("get", download_url)
    with open(download_file_path, "wb") as f:
        f.write(r.content)
    return download_file_path


def _request_with_retry(method: Text, url: Text, **params) -> requests.Response:
    """Wrapper around requests with Session to retry in case it fails

    Args:
        method (Text): HTTP method, such as 'GET' or 'HEAD'.
        url (Text): The URL of the resource to fetch.
        **params: Params to pass to request function

    Returns:
        requests.Response: Response object of the request
    """
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    response = session.request(method=method.upper(), url=url, **params)
    return response


def download_data(url_link, local_filename=None):
    if local_filename is None:
        local_filename = url_link.split("/")[-1]
    with requests.get(url_link, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                f.write(chunk)
    return local_filename


def upload_data(
    file_name: Union[Text, Path],
    tags: Optional[List[Text]] = None,
    license: Optional[License] = None,
    is_temp: bool = True,
    content_type: Text = "text/csv",
    content_encoding: Optional[Text] = None,
    nattempts: int = 2,
):
    """Upload files to S3 with pre-signed URLs

    Args:
        file_name (Union[Text, Path]): local path of file to be uploaded
        tags (List[Text], optional): tags of the file
        license (License, optional): the license for the file
        is_temp (bool): specify if the file that will be upload is a temporary file
        content_type (Text, optional): Type of content. Defaults to "text/csv".
        content_encoding (Text, optional): Content encoding. Defaults to None.
        nattempts (int, optional): Number of attempts for diminish the risk of exceptions. Defaults to 2.

    Reference:
        https://python.plainenglish.io/upload-files-to-aws-s3-using-pre-signed-urls-in-python-d3c2fcab1b41

    Returns:
        URL: s3 path
    """
    try:
        # Get pre-signed URL
        # URL of aiXplain service which returns a pre-signed URL to onboard the file
        if is_temp is True:
            url = urljoin(config.BACKEND_URL, "sdk/file/upload/temp-url")
            payload = {"contentType": content_type, "originalName": os.path.basename(file_name)}
        else:
            url = urljoin(config.BACKEND_URL, "sdk/file/upload-url")
            if tags is None:
                tags = []
            payload = {"contentType": content_type, "originalName": file_name, "tags": ",".join(tags), "license": license.value}

        if config.AIXPLAIN_API_KEY != "":
            team_key = config.AIXPLAIN_API_KEY
            headers = {"x-aixplain-key": team_key}
        else:
            team_key = config.TEAM_API_KEY
            headers = {"Authorization": "token " + team_key}

        r = _request_with_retry("post", url, headers=headers, data=payload)
        response = r.json()
        path = response["key"]
        # Upload data
        presigned_url = response["uploadUrl"]  # pre-signed URL
        headers = {"Content-Type": content_type}
        if content_encoding is not None:
            headers["Content-Encoding"] = content_encoding
        payload = open(file_name, "rb").read()
        # saving the file into the pre-signed URL
        r = _request_with_retry("put", presigned_url, headers=headers, data=payload)

        # if the process fail, try one more
        if r.status_code != 200:
            if nattempts > 0:
                return upload_data(file_name, content_type, content_encoding, nattempts - 1)
            else:
                raise Exception("File Uploading Error: Failure on Uploading to S3.")
        bucket_name = re.findall(r"https://(.*?).s3.amazonaws.com", presigned_url)[0]
        s3_link = f"s3://{bucket_name}/{path}"
        return s3_link
    except Exception as e:
        if nattempts > 0:
            return upload_data(file_name, content_type, content_encoding, nattempts - 1)
        else:
            raise Exception("File Uploading Error: Failure on Uploading to S3.")


def s3_to_csv(s3_url: Text, aws_credentials: Dict) -> Text:
    """Convert s3 url to a csv file and download the file in `download_path`

    Args:
        s3_url (Text): s3 url

    Returns:
        Path: path to csv file
    """
    try:
        import boto3
        from botocore.exceptions import NoCredentialsError
    except ModuleNotFoundError:
        raise Exception(
            "boto3 is not currently installed in your project environment. Please try installing it using pip:\n\npip install boto3"
        )
    url = urlparse(s3_url)
    if url.scheme != "s3":
        raise Exception("the url is not an s3 url")
    bucket_name = url.netloc
    prefix = url.path[1:]
    try:
        aws_access_key_id = aws_credentials.get("AWS_ACCESS_KEY_ID") or os.getenv("AWS_ACCESS_KEY_ID")
        aws_secret_access_key = aws_credentials.get("AWS_SECRET_ACCESS_KEY") or os.getenv("AWS_SECRET_ACCESS_KEY")
        s3 = boto3.client("s3", aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
        response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
    except NoCredentialsError as e:
        raise Exception(
            "to use the s3 bucket option you need to set the right AWS credentials [AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]"
        )
    except Exception as e:
        raise Exception("the bucket you are trying to use does not exist")

    try:
        files = []

        if "Contents" in response:
            for obj in response["Contents"]:
                files.append(obj["Key"])

        # check if no files where found
        if not files:
            raise Exception(f"ERROR No files were found => bucket name: {bucket_name}, prefix: {prefix}")

        response2 = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix, Delimiter="/")
        if "CommonPrefixes" in response2:
            there_are_folders = bool(response2["CommonPrefixes"])
        else:
            there_are_folders = False

        # check if there are folder or if the files in the root of the url, and reformart the paths
        if there_are_folders:
            data = defaultdict(list)
            for path in files:
                directory = path.rsplit("/", 2)[-2]
                data[directory].append(f"s3://{bucket_name}/{path}")

            # validate all the folders have the same length
            first_key = list(data.keys())[0]
            main_len = len(data[first_key])
            if any(main_len != len(val) for val in data.values()):
                raise Exception("all the directories should have the same number of files")

            # validate that the names of the files are the same in all the list
            if len(data.keys()) > 1:
                for i in range(main_len):
                    main_file_name = Path(data[first_key][i]).stem
                    for val in data.values():
                        if Path(val[i]).stem != main_file_name:
                            raise Exception(f"all the files in different directories should have the same prefix")

        elif prefix == "":
            raise Exception(f"ERROR the files can't be at the root of the bucket ")
        else:
            data = {prefix: [f"s3://{bucket_name}/{file}" for file in files]}

        # create DataFrame and convert it to csv
        df = DataFrame(data)
        output_path = f"{uuid4()}.csv"
        df.to_csv(output_path, index=False)

        return output_path
    except (Exception, ValueError) as e:
        raise Exception(e)



__author__ = "shreyassharma"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: December 1st 2022
Description:
    Metric Factory Class
"""

import logging
import os
from typing import List, Optional
from aixplain.modules import Metric
from aixplain.utils import config
from aixplain.utils.file_utils import _request_with_retry
from typing import Dict, Text
from urllib.parse import urljoin
from warnings import warn


class MetricFactory:
    """A static class for creating and exploring Metric Objects.

    Attributes:
        backend_url (str): The URL for the backend.
    """

    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def _create_metric_from_response(cls, response: Dict) -> Metric:
        """Converts response Json to 'Metric' object

        Args:
            response (Dict): Json from API

        Returns:
            Metric: Coverted 'Metric' object
        """
        return Metric(
            id=response["id"],
            name=response["name"],
            supplier=response["supplier"],
            is_reference_required=response["referenceRequired"],
            is_source_required=response["sourceRequired"],
            cost=response["normalizedPrice"],
            function=response["function"],
        )

    @classmethod
    def get(cls, metric_id: Text) -> Metric:
        """Create a 'Metric' object from metric id

        Args:
            model_id (Text): Model ID of required metric.

        Returns:
            Metric: Created 'Metric' object
        """

        resp, status_code = None, 200
        try:
            url = urljoin(cls.backend_url, f"sdk/metrics/{metric_id}")
            if cls.aixplain_key != "":
                headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
            else:
                headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
            logging.info(f"Start service for GET Metric  - {url} - {headers}")
            r = _request_with_retry("get", url, headers=headers)
            resp = r.json()
            metric = cls._create_metric_from_response(resp)
        except Exception as e:
            if resp is not None and "statusCode" in resp:
                status_code = resp["statusCode"]
                message = resp["message"]
                message = f"Metric Creation: Status {status_code} - {message}"
            else:
                message = "Metric Creation: Unspecified Error"
            logging.error(f"Metric Creation Failed: {e}")
            raise Exception(f"Status {status_code}: {message}")
        return metric

    @classmethod
    def list(
        cls,
        model_id: Text = None,
        is_source_required: Optional[bool] = None,
        is_reference_required: Optional[bool] = None,
        page_number: int = 0,
        page_size: int = 20,
    ) -> List[Metric]:
        """Get list of supported metrics for the given filters

        Args:
            model_id (Text, optional): ID of model for which metric is to be used. Defaults to None.
            is_source_required (bool, optional): Should the metric use source. Defaults to None.
            is_reference_required (bool, optional): Should the metric use reference. Defaults to None.
            page_number (int, optional): page number. Defaults to 0.
            page_size (int, optional): page size. Defaults to 20.

        Returns:
            List[Metric]: List of supported metrics
        """
        try:
            url = urljoin(cls.backend_url, f"sdk/metrics")
            filter_params = {}
            if model_id is not None:
                filter_params["modelId"] = model_id
            if is_source_required is not None:
                filter_params["sourceRequired"] = 1 if is_source_required else 0
            if is_reference_required is not None:
                filter_params["referenceRequired"] = 1 if is_reference_required else 0

            if cls.aixplain_key != "":
                headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
            else:
                headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
            r = _request_with_retry("get", url, headers=headers, params=filter_params)
            resp = r.json()
            logging.info(f"Listing Metrics: Status of getting metrics: {resp}")
            all_metrics = resp["results"]
            starting_model_index_overall = page_number * page_size
            ending_model_index_overall = starting_model_index_overall + page_size - 1
            filtered_metrics = all_metrics[starting_model_index_overall : ending_model_index_overall + 1]
            total = len(filtered_metrics)
            metric_list = [cls._create_metric_from_response(metric_info_json) for metric_info_json in filtered_metrics]
            return {
                "results": metric_list,
                "page_total": min(page_size, len(metric_list)),
                "page_number": page_number,
                "total": total,
            }
        except Exception as e:
            error_message = f"Listing Metrics: Error in getting metrics: {e}"
            logging.error(error_message, exc_info=True)
            return []



__author__ = "lucaspavanelli"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: September 1st 2022
Description:
    Model Class
"""
import time
import json
import logging
import traceback
from aixplain.enums import Supplier, Function
from aixplain.modules.asset import Asset
from aixplain.utils import config
from urllib.parse import urljoin
from aixplain.utils.file_utils import _request_with_retry
from typing import Union, Optional, Text, Dict
from datetime import datetime


class Model(Asset):
    """This is ready-to-use AI model. This model can be run in both synchronous and asynchronous manner.

    Attributes:
        id (Text): ID of the Model
        name (Text): Name of the Model
        description (Text, optional): description of the model. Defaults to "".
        api_key (Text, optional): API key of the Model. Defaults to None.
        url (Text, optional): endpoint of the model. Defaults to config.MODELS_RUN_URL.
        supplier (Union[Dict, Text, Supplier, int], optional): supplier of the asset. Defaults to "aiXplain".
        version (Text, optional): version of the model. Defaults to "1.0".
        function (Function, optional): model AI function. Defaults to None.
        url (str): URL to run the model.
        backend_url (str): URL of the backend.
        pricing (Dict, optional): model price. Defaults to None.
        **additional_info: Any additional Model info to be saved
        input_params (Dict, optional): input parameters for the function.
        output_params (Dict, optional): output parameters for the function.
    """

    def __init__(
        self,
        id: Text,
        name: Text,
        description: Text = "",
        api_key: Text = config.TEAM_API_KEY,
        supplier: Union[Dict, Text, Supplier, int] = "aiXplain",
        version: Optional[Text] = None,
        function: Optional[Function] = None,
        is_subscribed: bool = False,
        cost: Optional[Dict] = None,
        created_at: Optional[datetime] = None,
        input_params: Optional[Dict] = None,
        output_params: Optional[Dict] = None,
        **additional_info,
    ) -> None:
        """Model Init

        Args:
            id (Text): ID of the Model
            name (Text): Name of the Model
            description (Text, optional): description of the model. Defaults to "".
            api_key (Text, optional): API key of the Model. Defaults to None.
            supplier (Union[Dict, Text, Supplier, int], optional): supplier of the asset. Defaults to "aiXplain".
            version (Text, optional): version of the model. Defaults to "1.0".
            function (Text, optional): model AI function. Defaults to None.
            is_subscribed (bool, optional): Is the user subscribed. Defaults to False.
            cost (Dict, optional): model price. Defaults to None.
            **additional_info: Any additional Model info to be saved
        """
        super().__init__(id, name, description, supplier, version, cost=cost)
        self.api_key = api_key
        self.additional_info = additional_info
        self.url = config.MODELS_RUN_URL
        self.backend_url = config.BACKEND_URL
        self.function = function
        self.is_subscribed = is_subscribed
        self.created_at = created_at
        self.input_params = input_params
        self.output_params = output_params

    def to_dict(self) -> Dict:
        """Get the model info as a Dictionary

        Returns:
            Dict: Model Information
        """
        clean_additional_info = {k: v for k, v in self.additional_info.items() if v is not None}
        return {
            "id": self.id,
            "name": self.name,
            "supplier": self.supplier,
            "additional_info": clean_additional_info,
            "input_params": self.input_params,
            "output_params": self.output_params,
        }

    def __repr__(self):
        try:
            return f"<Model: {self.name} by {self.supplier['name']}>"
        except Exception:
            return f"<Model: {self.name} by {self.supplier}>"

    def sync_poll(self, poll_url: Text, name: Text = "model_process", wait_time: float = 0.5, timeout: float = 300) -> Dict:
        """Keeps polling the platform to check whether an asynchronous call is done.

        Args:
            poll_url (Text): polling URL
            name (Text, optional): ID given to a call. Defaults to "model_process".
            wait_time (float, optional): wait time in seconds between polling calls. Defaults to 0.5.
            timeout (float, optional): total polling time. Defaults to 300.

        Returns:
            Dict: response obtained by polling call
        """
        logging.info(f"Polling for Model: Start polling for {name}")
        start, end = time.time(), time.time()
        # keep wait time as 0.2 seconds the minimum
        wait_time = max(wait_time, 0.2)
        completed = False
        response_body = {"status": "FAILED", "completed": False}
        while not completed and (end - start) < timeout:
            try:
                response_body = self.poll(poll_url, name=name)
                completed = response_body["completed"]

                end = time.time()
                if completed is False:
                    time.sleep(wait_time)
                    if wait_time < 60:
                        wait_time *= 1.1
            except Exception as e:
                response_body = {"status": "FAILED", "completed": False, "error": "No response from the service."}
                logging.error(f"Polling for Model: polling for {name}: {e}")
                break
        if response_body["completed"] is True:
            logging.info(f"Polling for Model: Final status of polling for {name}: {response_body}")
        else:
            response_body["status"] = "FAILED"
            logging.error(
                f"Polling for Model: Final status of polling for {name}: No response in {timeout} seconds - {response_body}"
            )
        return response_body

    def poll(self, poll_url: Text, name: Text = "model_process") -> Dict:
        """Poll the platform to check whether an asynchronous call is done.

        Args:
            poll_url (Text): polling
            name (Text, optional): ID given to a call. Defaults to "model_process".

        Returns:
            Dict: response obtained by polling call
        """
        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}
        r = _request_with_retry("get", poll_url, headers=headers)
        try:
            resp = r.json()
            if resp["completed"] is True:
                resp["status"] = "SUCCESS"
                if "error" in resp or "supplierError" in resp:
                    resp["status"] = "FAILED"
            else:
                resp["status"] = "IN_PROGRESS"
            logging.debug(f"Single Poll for Model: Status of polling for {name}: {resp}")
        except Exception as e:
            resp = {"status": "FAILED"}
            logging.error(f"Single Poll for Model: Error of polling for {name}: {e}")
        return resp

    def run(
        self,
        data: Union[Text, Dict],
        name: Text = "model_process",
        timeout: float = 300,
        parameters: Dict = {},
        wait_time: float = 0.5,
    ) -> Dict:
        """Runs a model call.

        Args:
            data (Union[Text, Dict]): link to the input data
            name (Text, optional): ID given to a call. Defaults to "model_process".
            timeout (float, optional): total polling time. Defaults to 300.
            parameters (Dict, optional): optional parameters to the model. Defaults to "{}".
            wait_time (float, optional): wait time in seconds between polling calls. Defaults to 0.5.

        Returns:
            Dict: parsed output from model
        """
        start = time.time()
        try:
            response = self.run_async(data, name=name, parameters=parameters)
            if response["status"] == "FAILED":
                end = time.time()
                response["elapsed_time"] = end - start
                return response
            poll_url = response["url"]
            end = time.time()
            response = self.sync_poll(poll_url, name=name, timeout=timeout, wait_time=wait_time)
            return response
        except Exception as e:
            msg = f"Error in request for {name} - {traceback.format_exc()}"
            logging.error(f"Model Run: Error in running for {name}: {e}")
            end = time.time()
            return {"status": "FAILED", "error": msg, "elapsed_time": end - start}

    def run_async(self, data: Union[Text, Dict], name: Text = "model_process", parameters: Dict = {}) -> Dict:
        """Runs asynchronously a model call.

        Args:
            data (Union[Text, Dict]): link to the input data
            name (Text, optional): ID given to a call. Defaults to "model_process".
            parameters (Dict, optional): optional parameters to the model. Defaults to "{}".

        Returns:
            dict: polling URL in response
        """
        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}
        from aixplain.factories.file_factory import FileFactory

        data = FileFactory.to_link(data)
        if isinstance(data, dict):
            payload = data
        else:
            try:
                payload = json.loads(data)
                if isinstance(payload, dict) is False:
                    if isinstance(payload, int) is True or isinstance(payload, float) is True:
                        payload = str(payload)
                    payload = {"data": payload}
            except Exception:
                payload = {"data": data}
        payload.update(parameters)
        payload = json.dumps(payload)

        call_url = f"{self.url}/{self.id}"
        r = _request_with_retry("post", call_url, headers=headers, data=payload)
        logging.info(f"Model Run Async: Start service for {name} - {self.url} - {payload} - {headers}")

        resp = None
        try:
            if 200 <= r.status_code < 300:
                resp = r.json()
                logging.info(f"Result of request for {name} - {r.status_code} - {resp}")
                poll_url = resp["data"]
                response = {"status": "IN_PROGRESS", "url": poll_url}
            else:
                if r.status_code == 401:
                    error = "Unauthorized API key: Please verify the spelling of the API key and its current validity."
                elif 460 <= r.status_code < 470:
                    error = "Subscription-related error: Please ensure that your subscription is active and has not expired."
                elif 470 <= r.status_code < 480:
                    error = "Billing-related error: Please ensure you have enough credits to run this model. "
                elif 480 <= r.status_code < 490:
                    error = "Supplier-related error: Please ensure that the selected supplier provides the model you are trying to access."
                elif 490 <= r.status_code < 500:
                    error = "Validation-related error: Please ensure all required fields are provided and correctly formatted."
                else:
                    status_code = str(r.status_code)
                    error = (
                        f"Status {status_code}: Unspecified error: An unspecified error occurred while processing your request."
                    )
                response = {"status": "FAILED", "error_message": error}
                logging.error(f"Error in request for {name} - {r.status_code}: {error}")
        except Exception:
            response = {"status": "FAILED"}
            msg = f"Error in request for {name} - {traceback.format_exc()}"
            logging.error(f"Model Run Async: Error in running for {name}: {resp}")
            if resp is not None:
                response["error"] = msg
        return response

    def check_finetune_status(self, after_epoch: Optional[int] = None):
        """Check the status of the FineTune model.

        Args:
            after_epoch (Optional[int], optional): status after a given epoch. Defaults to None.

        Raises:
            Exception: If the 'TEAM_API_KEY' is not provided.

        Returns:
            FinetuneStatus: The status of the FineTune model.
        """
        from aixplain.enums.asset_status import AssetStatus
        from aixplain.modules.finetune.status import FinetuneStatus

        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}
        resp = None
        try:
            url = urljoin(self.backend_url, f"sdk/finetune/{self.id}/ml-logs")
            logging.info(f"Start service for GET Check FineTune status Model  - {url} - {headers}")
            r = _request_with_retry("get", url, headers=headers)
            resp = r.json()
            finetune_status = AssetStatus(resp["finetuneStatus"])
            model_status = AssetStatus(resp["modelStatus"])
            logs = sorted(resp["logs"], key=lambda x: float(x["epoch"]))
            target_epoch = None
            if after_epoch is not None:
                logs = [log for log in logs if float(log["epoch"]) > after_epoch]
                if len(logs) > 0:
                    target_epoch = float(logs[0]["epoch"])
            elif len(logs) > 0:
                target_epoch = float(logs[-1]["epoch"])
            if target_epoch is not None:
                log = None
                for log_ in logs:
                    if int(log_["epoch"]) == target_epoch:
                        if log is None:
                            log = log_
                        else:
                            if log_["trainLoss"] is not None:
                                log["trainLoss"] = log_["trainLoss"]
                            if log_["evalLoss"] is not None:
                                log["evalLoss"] = log_["evalLoss"]
                status = FinetuneStatus(
                    status=finetune_status,
                    model_status=model_status,
                    epoch=float(log["epoch"]) if "epoch" in log and log["epoch"] is not None else None,
                    training_loss=float(log["trainLoss"]) if "trainLoss" in log and log["trainLoss"] is not None else None,
                    validation_loss=float(log["evalLoss"]) if "evalLoss" in log and log["evalLoss"] is not None else None,
                )
            else:
                status = FinetuneStatus(
                    status=finetune_status,
                    model_status=model_status,
                )

            logging.info(f"Response for GET Check FineTune status Model - Id {self.id} / Status {status.status.value}.")
            return status
        except Exception:
            message = ""
            if resp is not None and "statusCode" in resp:
                status_code = resp["statusCode"]
                message = resp["message"]
                message = f"Status {status_code} - {message}"
            error_message = f"Check FineTune status Model: Error {message}"
            logging.exception(error_message)

    def delete(self) -> None:
        """Delete Model service"""
        try:
            url = urljoin(self.backend_url, f"sdk/models/{self.id}")
            headers = {"Authorization": f"Token {self.api_key}", "Content-Type": "application/json"}
            logging.info(f"Start service for DELETE Model  - {url} - {headers}")
            r = _request_with_retry("delete", url, headers=headers)
            if r.status_code != 200:
                raise Exception()
        except Exception:
            message = "Model Deletion Error: Make sure the model exists and you are the owner."
            logging.error(message)
            raise Exception(f"{message}")
